
/*
 * blueMail offline mail reader
 * class definitions for bmail

 Copyright (c) 1996 Toth Istvan <stoty@vma.bme.hu>
 Copyright (c) 1998 William McBrine <wmcbrine@clark.net>
 Copyright (c) 2002 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */

#ifndef BMAIL_H
#define BMAIL_H


#include <time.h>
#include <stdio.h>
#include <sys/types.h>
#include "resource.h"
#include "service.h"


// area types
#define A_COLLECTION 0x0001   // collection of messages from other areas
#define A_REPLYAREA  0x0002   // the reply area
#define A_SUBSCRIBED 0x0004   // area is active, i.e. has been subscribed
#define A_FORCED     0x0008   // area may not be unsubscribed
#define A_NETMAIL    0x0010   // netmail area (not an echo area)
#define A_INTERNET   0x0020   // area is internet style (news, email)
#define A_READONLY   0x0040   // no posting allowed
#define A_ALIAS      0x0080   // use alias name in the this area
#define A_ALLOWPRIV  0x0100   // allow use of private flag in this echo area

// message types (Blue Wave compatible)
#define PRIVATE       0x0001   // for addressee only
#define CRASH         0x0002   // high priority mail
#define RECEIVED      0x0004   // message received
#define SENT          0x0008   // message sent
#define FILEATTACHED  0x0010   // message with attached file
#define INTRANSIT     0x0020   // message to be forwarded (to others)
#define ORPHAN        0x0040   // message destination unknown
#define KILL          0x0080   // kill (delete) after sending
#define LOCAL         0x0100   // message originated here
#define HOLD          0x0200   // hold for pickup, don't send
#define IMMEDIATE     0x0400   // send message NOW
#define FILEREQUEST   0x0800   // file request
#define DIRECT        0x1000   // send direct, no routing
#define UNUSED_M1              // unused
#define UNUSED_M2              // unused
#define FILEUPDATEREQ 0x8000   // updated file request

// message type strings
#define CRASH_TXT         "Crash"
#define RECEIVED_TXT      "Rcvd"
#define SENT_TXT          "Sent"
#define FILEATTACHED_TXT  "w/File"
#define INTRANSIT_TXT     "Trans"
#define ORPHAN_TXT        "Orphan"
#define KILL_TXT          "K/Sent"
#define LOCAL_TXT         "Local"
#define HOLD_TXT          "Hold"
#define IMMEDIATE_TXT     "Imm"
#define FILEREQUEST_TXT   "F/Req"
#define DIRECT_TXT        "Direct"
#define FILEUPDATEREQ_TXT "U/Req"

// message status flags (Blue Wave compatible)
#define MSF_READ     0x0001
#define MSF_REPLIED  0x0002
#define MSF_PERSONAL 0x0004
#define MSF_TAGGED   0x0008   // unused
#define MSF_SAVED    0x0010   // unused
#define MSF_PRINTED  0x0020   // unused
#define MSF_MARKED   0x0F00   // bmail switches between 0 (false) and 0x0F00
                              // (true) but accepts any bit set (true) to be
                              // interoperable with the four Blue Wave marks
                              // (which is no problem as bmail has only one)
// for file_list::sort()
#define FL_SORT_BY_NAME   0
#define FL_SORT_BY_DATE   1   // unused
#define FL_SORT_BY_NEWEST 2

// for letter_list::sort()
#define LL_SORT_BY_SUBJ   0
#define LL_SORT_BY_MSGNUM 1
#define LL_SORT_FIRSTTYPE LL_SORT_BY_SUBJ
#define LL_SORT_LASTTYPE  LL_SORT_BY_MSGNUM

// address lengths
#define FIDOADDRLEN  23
#define OTHERADDRLEN 99


// return types for saveLastread()
enum lrtype {LR_OK, LR_ERROR, LR_NORETRY};


// message body access structure
struct bodytype
{
  long pointer;
  long msglen;
};


// necessary forward references
class bmail;
class driver_list;
class main_driver;
class reader;


class net_address
{
  private:
    unsigned int zone, net, node, point;
    char *otherAddr;

    void copy(net_address &);

  public:
    net_address();
    net_address(net_address &);
    ~net_address();
    net_address &operator =(net_address &);
    void set(unsigned int, unsigned int, unsigned int, unsigned int);
    unsigned int getZone() const;
    unsigned int getNet() const;
    unsigned int getNode() const;
    unsigned int getPoint() const;
    void set(const char *);
    char *get(bool = false) const;
};


class file_header
{
  private:
    char *name;
    time_t date;
    off_t size;

  public:
    file_header *next;

    file_header(char *, time_t, off_t);
    ~file_header();
    const char *getName() const;
    time_t getDate() const;
    off_t getSize() const;
};


class file_list
{
  private:
    const char *dirname;
    file_header **files;
    int noOfFiles, activeFile;

    friend int fsortbyname(const void *, const void *);
    friend int fsortbydate(const void *, const void *);
    friend int fsortbynewest(const void *, const void *);

  public:
    file_list(const char *);
    ~file_list();
    const char *getDir() const;
    void sort(int);
    int getNoOfFiles() const;
    void gotoFile(int);
    const char *getName() const;
    time_t getDate() const;
    off_t getSize() const;
    const char *exists(const char *);
    const char *getNext(const char *);
    FILE *ftryopen(const char *, const char *);
    void add(const char **, const char *, int *);
};


class packet_stat
{
  private:
    resource *ro;

    struct pktstat
    {
      time_t pktDate;
      int msgsTotal;
      int msgsUnread;
      bool used;
      pktstat *next;
    } *anchor, *last;

    void newStat(pktstat *);

  public:
    packet_stat(bmail *);
    ~packet_stat();
    void addStat(time_t, int, int);
    bool getStat(time_t, int *, int *) const;
};


class area_header
{
  private:
    bmail *bm;
    int area, flags, noOfLetters, noOfPersonal, maxFromToLen, maxSubjLen;
    const char *number, *shortname, *title, *type;
    main_driver *driver;

  public:
    area_header(bmail *, int, const char *, const char *, const char *,
                const char *, int, int, int, int, int);
    ~area_header();
    const char *getNumber() const;
    const char *getShortName() const;
    const char *getTitle() const;
    const char *getAreaType() const;
    bool isCollection() const;
    bool isReplyArea() const;
    bool isSubscribed() const;
    bool isForced() const;
    bool useAlias() const;
    bool isNetmail() const;
    bool isInternet() const;
    bool isReadonly() const;
    bool hasTo() const;
    bool allowPrivate() const;
    int getNoOfLetters() const;
    int getNoOfUnread();
    int getNoOfPersonal() const;
    int getMaxFromToLen() const;
    int getMaxSubjLen() const;
    int supportedMSF() const;
};


class area_list
{
  private:
    bmail *bm;
    int noOfAreas, currentArea, *activeHeader, noOfActive;
    bool shortlist, *filter;
    area_header **areaHeader;

    int checkArea(int) const;
    void refreshReplyArea();

  public:
    area_list(bmail *);
    ~area_list();
    void relist(bool);
    const char *getNumber() const;
    const char *getShortName(int = -1) const;
    const char *getTitle(int = -1) const;
    const char *getAreaType() const;
    bool isCollection() const;
    bool isReplyArea() const;
    bool isSubscribed() const;
    bool isForced() const;
    bool useAlias() const;
    bool isNetmail(int = -1) const;
    bool isInternet(int = -1) const;
    bool isReadonly() const;
    bool hasTo() const;
    bool allowPrivate() const;
    int getNoOfLetters() const;
    int getNoOfUnread() const;
    int getNoOfPersonal() const;
    int getMaxFromToLen() const;
    int getMaxSubjLen() const;
    int supportedMSF() const;
    int getNoOfAreas() const;
    int getNoOfActive() const;
    int getAreaNo() const;
    int getActive();
    void gotoArea(int);
    void gotoActive(int);
    bool findNetmail(int *) const;
    void getLetterList();
    void enterLetter(int, const char *, const char *, const char *, int, int,
                     net_address *, const char *, long);
    void killLetter(int);
    const char *newLetterBody(int, long);
    bool repliesOK();
    bool makeReply();
    void updateCollectionStatus();
    void setFilter(bool *);
};


class letter_header
{
  private:
    driver_list *dl;
    resource *ro;
    main_driver *driver;
    reader *read;
    char *from, *to, *subject, *date;
    int LetterID, msgNum, replyTo, AreaID, flags;
    long length;
    bool latin1;
    net_address netAddr;

  public:
    letter_header(bmail *, int, const char *, const char *, const char *,
                  const char *, int, int, int, int, bool, net_address *,
                  main_driver *);
    ~letter_header();
    int getLetterID() const;
    const char *getFrom() const;
    void setFrom(const char *);
    const char *getTo() const;
    void setTo(const char *);
    const char *getSubject() const;
    void setSubject(const char *);
    const char *getDate() const;
    int getMsgNum() const;
    int getReplyTo() const;
    int getAreaID() const;
    int getFlags() const;
    bool isPrivate() const;
    char *Flags() const;
    bool isLatin1() const;
    void setLatin1(bool);
    net_address &getNetAddr();
    main_driver *getDriver() const;
    bool isPersonal() const;
    bool isRead() const;
    void setRead();
    void setMarked();
    int getStatus() const;
    void setStatus(int);
    const char *getBody();
};


class letter_list
{
  private:
    resource *ro;
    main_driver *driver;
    reader *read;
    int area, areaType, noOfLetters, currentLetter, *activeHeader, noOfActive;
    bool shortlist, *filter;
    letter_header **letterHeader;

    void init();
    void cleanup();

    friend int lsortbysubj(const void *, const void *);
    friend int lsortbymsgnum(const void *, const void *);

  public:
    letter_list(bmail *, int, int);
    ~letter_list();
    void sort(int);
    void relist(bool);
    const char *getFrom() const;
    const char *getTo() const;
    const char *getSubject() const;
    const char *getDate() const;
    int getMsgNum() const;
    int getReplyTo() const;
    int getAreaID() const;
    int getFlags() const;
    bool isPrivate() const;
    char *Flags() const;
    bool isLatin1() const;
    net_address &getNetAddr();
    bool isPersonal() const;
    bool isRead() const;
    void setRead();
    int getStatus() const;
    void setStatus(int);
    const char *getBody() const;
    int getNoOfLetters() const;
    int getNoOfActive() const;
    void gotoLetter(int);
    void gotoActive(int);
    int getCurrent() const;
    int getActive() const;
    void refreshLetters();
    void setFilter(bool *);
};


class main_driver
{
  public:
    virtual ~main_driver();
    virtual const char *init() = 0;
    virtual int getNoOfAreas() = 0;
    virtual area_header *getNextArea() = 0;
    virtual void selectArea(int) = 0;
    virtual int getNoOfLetters() = 0;
    virtual letter_header *getNextLetter() = 0;
    virtual const char *getBody(int, int, letter_header *) = 0;
    virtual void resetLetters() = 0;
    virtual void initMSF(int **) = 0;
    virtual int readMSF(int **) = 0;
    virtual const char *saveMSF(int **) = 0;
    virtual bool isLatin1() = 0;
    virtual const char **getBulletins() = 0;
};


class reply_driver : public main_driver
{
  private:
    virtual void replyInf() = 0;

  public:
    virtual ~reply_driver();
    virtual void enterLetter(letter_header *, const char *, long) = 0;
    virtual void killLetter(int) = 0;
    virtual const char *newLetterBody(int, long) = 0;
    virtual area_header *getArea() = 0;
    virtual bool makeReply() = 0;
};


class reader
{
  public:
    virtual ~reader();
    virtual int supportedMSF() = 0;
    virtual void setRead(int, int, bool) = 0;
    virtual bool isRead(int, int) const = 0;
    virtual void setStatus(int, int, int) = 0;
    virtual int getStatus(int, int) const = 0;
    virtual int getNoOfUnread(int) const = 0;
    virtual int getNoOfPersonal(int) const = 0;
    virtual const char *saveMSF() = 0;
};


class main_reader : public reader
{
  private:
    bmail *bm;
    main_driver *driver;
    int noOfAreas, *noOfLetters, **msgstat, msfs;

  public:
    main_reader(bmail *, main_driver *);
    ~main_reader();
    int supportedMSF();
    void setRead(int, int, bool);
    bool isRead(int, int) const;
    void setStatus(int, int, int);
    int getStatus(int, int) const;
    int getNoOfUnread(int) const;
    int getNoOfPersonal(int) const;
    const char *saveMSF();
};


class reply_reader: public reader
{
  public:
    reply_reader(bmail *, main_driver *);
    ~reply_reader();
    int supportedMSF();
    void setRead(int, int, bool);
    bool isRead(int, int) const ;
    void setStatus(int, int, int);
    int getStatus(int, int) const;
    int getNoOfUnread(int) const;
    int getNoOfPersonal(int) const;
    const char *saveMSF();
};


class driver_list
{
  private:
    struct
    {
      main_driver *driver;
      reader *read;
      int offset;
    } driverList[2];     // 0: main_driver/main_reader   (must exist)
                         // 1: reply_driver/reply_reader (may exist)
    int noOfDrivers, attributes;
    char *envVar;
    const char *driverID;
    char driverEnv[21];

  public:
    driver_list(bmail *);
    ~driver_list();
    const char *init(bmail *);
    int getNoOfDrivers() const;
    bool canReply() const;
    main_driver *getDriver(int) const;
    main_driver *getMainDriver() const;
    reply_driver *getReplyDriver() const;
    reader *getReadObject(main_driver *) const;
    int getOffset(main_driver *) const;
    bool hasPersonal() const;
    bool useTearline() const;
    bool allowQuotedPrintable() const;
    void toggleQuotedPrintable();
};


class bmail
{
  private:
    char bm_version[41];
    char select_error[80];
    time_t lastPktMtime;
    servicetype serviceType;

    bool loadDriver(const char *);

  public:
    resource *resourceObject;
    services *service;
    file_list *fileList;
    driver_list *driverList;
    area_list *areaList;
    letter_list *letterList;

    bmail();
    ~bmail();
    void Delete();
    bool selectPacket(const char *);
    bool selectFileDB(const char *, bool);
    bool selectReply(const char *);
    bool selectDemo();
    servicetype getServiceType() const;
    lrtype saveLastread();
    time_t getLastPktMtime() const;
    const char *version() const;
    bool isLatin1() const;
    const char **getBulletins() const;
    const char *getSelectError() const;
};


#endif
