/*
 * blueMail offline mail reader
 * routines common to more than one driver

 Copyright (c) 2002 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */


#include <ctype.h>
#include <string.h>
#include "alldrv.h"
#ifndef BIG_ENDIAN
  #define BIG_ENDIAN
#endif
#include "bluewave.h"
#include "../../common/auxil.h"
#include "../../common/error.h"
#include "../../common/mysystem.h"


alldriver::alldriver (bmail *bm, main_driver *driver)
{
  this->bm = bm;
  this->driver = driver;
  from = faddr = to = subject = date = NULL;
}


alldriver::~alldriver ()
{
  delete[] date;
  delete[] subject;
  delete[] to;
  delete[] faddr;
  delete[] from;
}


// determine character set according to FidoNet Standards Proposal 0054
// (I need kludge 2, too. Hopefully, nobody will notice or complain.)
void alldriver::fsc0054 (const char *message, char kludge, bool *latin1)
{
  const char *found;
  static char chrs[] = " CHRS: ";
  static char charset[] = " CHARSET: ";

  chrs[0] = charset[0] = kludge;

  if ((found = strstr(message, chrs))) found += strlen(chrs);
  if (!found && (found = strstr(message, charset))) found += strlen(charset);
  if (found)
  {
    if (strncasecmp(found, "LATIN-1", 7) == 0) *latin1 = true;
    if (strncasecmp(found, "IBMPC", 5) == 0) *latin1 = false;
  }
}


// read Blue Wave style XTI file
void alldriver::XTIread (int **msgstat, int offset, const char *pfname)
{
  FILE *xtifile;
  XTI_REC xti;

  if ((xtifile = bm->fileList->ftryopen(pfname, "rb")))
  {
    // skip collection areas
    for (int a = offset; a < driver->getNoOfAreas(); a++)
    {
      driver->selectArea(a);
      for (int l = 0; l < driver->getNoOfLetters(); l++)
      {
        fscanf(xtifile, "%c%c", &xti.flags, &xti.marks);
        // message status flags match XTI flags
        msgstat[a][l] = (xti.marks << 8) + xti.flags;
      }
    }
    fclose(xtifile);
  }
  else driver->initMSF(msgstat);
}


// write Blue Wave style XTI file
const char *alldriver::XTIsave (int **msgstat, int offset,
                                const char *name, bool extxti)
{
  FILE *xtifile;
  static char xtifname[13];
  XTI_REC xti;

  sprintf(xtifname, "%.8s%s", name, (extxti ? ".xti" : ""));

  if ((xtifile = fopen(xtifname, "wb")))
  {
    // skip collection areas
    for (int a = offset; a < driver->getNoOfAreas(); a++)
    {
      driver->selectArea(a);
      for (int l = 0; l < driver->getNoOfLetters(); l++)
      {
        // message status flags match XTI flags
        xti.flags = msgstat[a][l] & 0xFF;
        xti.marks = msgstat[a][l] >> 8;
        fprintf(xtifile, "%c%c", xti.flags, xti.marks);
      }
    }
    fclose(xtifile);
    return xtifname;
  }
  else return NULL;
}


char *alldriver::getFidoAddr (const char *msg)
{
  char *pos, *origin = strstr(msg, "\n * Origin: ");

  if (origin)
  {
    pos = origin + 1;

    // search end of line
    while (*pos && (*pos != '\n')) pos++;

    // search opening bracket
    while (pos > origin && *pos != '(') pos--;

    // according to FidoNet Standards Proposal 0007 the nodeid must follow
    // the opening bracket, but to handle violations, too, we skip non-digits
    while (*pos && (*pos != '\n') && !is_digit(*pos)) pos++;

    if (is_digit(*pos)) return pos;
  }

  return NULL;
}


area_header *alldriver::getReplyArea (const char *type, int noOfLetters)
{
  // REPLY_AREA has area number 0
  return new area_header(bm, 0, "REPLY", "REPLIES", "Letters written by you",
                         type, A_REPLYAREA | A_SUBSCRIBED | A_FORCED,
                         noOfLetters, 0, 0, 0);
}


file_list *alldriver::unpackReplies (const char *replyPacketName)
{
  char fname[MYMAXPATH];
  const char *replyWorkDir = bm->resourceObject->get(ReplyWorkDir);
  file_list *replies = NULL;

  mkfname(fname, bm->resourceObject->get(ReplyDir), replyPacketName);

  if (access(fname, R_OK | W_OK) == 0)
    if (bm->service->unpack(fname, replyWorkDir, false) == 0)
      replies = new file_list(replyWorkDir);

  return replies;
}


void alldriver::fidodate (char *date, unsigned int size, time_t time)
{
  static char fmt[] = "%d %b %y  %H:%M:%S";     // to avoid warning for %y

  strftime(date, size, fmt, localtime(&time));
}


void alldriver::mimeCharset (const char *buffer, bool *latin1)
{
  const char *found, *cset;
  size_t lenc;

  if (strncasecmp(buffer, "Content-Type: ", 14) == 0 || *buffer == '\t')
  {
    if ((found = stricstr(buffer, "charset=")))
    {
      found += 8;
      while (*found && (*found == ' ' || *found == '"')) found++;

      int i = 0;
      while ((cset = charsetmap[i++].charset))
      {
        lenc = strlen(cset);

        if ((strncasecmp(found, cset, lenc) == 0) && !is_digit(found[lenc]))
        {
          *latin1 = charsetmap[--i].isLatin1;
          break;
        }
      }
    }
  }
}


void alldriver::splitAddress (const char *na, int &name, int &n_len,
                                              int &addr, int &a_len)
{
  char *pos;

  n_len = a_len = 0;

  if ((pos = strchr(na, '(')))
  {
    name = pos - na + 1;
    addr = 0;
    a_len = pos - na - 1;
    while (*pos && (*pos != ')'))
    {
      n_len++;
      pos++;
    }
    n_len--;
  }
  else if ((pos = strchr(na, '<')))
  {
    name = 0;
    addr = pos - na + 1;
    n_len = addr - 2;
    while (*pos && (*pos != '>'))
    {
      a_len++;
      pos++;
    }
    a_len--;

    if (*na == '"')
    {
      name++;
      n_len--;
    }
    int end = n_len - 1;
    if (end >= 0 && na[name + end] == '"') n_len--;
  }
  else if ((pos = strchr(na, '@')))
  {
    name = addr = 0;
    a_len = strlen(na);
  }
  else
  {
    name = addr = 0;
    n_len = strlen(na);
  }

  if (n_len < 0) n_len = 0;
  if (a_len < 0) a_len = 0;
}


const char **alldriver::bulletin_list (const char file[][13], int count,
                                                              int type)
{
  file_list *fl = bm->fileList;
  int files = 0;
  const char **bulletins = new const char *[fl->getNoOfFiles() + 1];

  for (int i = 0; i < count; i++)
    if (file[i][0]) fl->add(bulletins, file[i], &files);

  if (type & BL_QWK)
  {
    fl->add(bulletins, "blt-*", &files);
    fl->add(bulletins, "session.txt", &files);
  }

  if (type & BL_NEWFILES) fl->add(bulletins, "newfiles.*", &files);

  bulletins[files] = NULL;

  return bulletins;
}


bool alldriver::isQP (const char *source)
{
  const char *found;
  static char qp[] = "\n CONTENT-TRANSFER-ENCODING: QUOTED-PRINTABLE\n";

  qp[1] = '\1';
  found = stricstr(source, qp);

  if (!found && stricstr(source, "\n\1CONTENT-TYPE: MULTIPART"))
  {
    qp[1] = '\n';
    found = stricstr(source, qp + 1);
  }

  return (found != NULL);
}


// parse a header conforming to RFC822 and RFC1036 stored in a file,
// set charset and return number of headerlines
long alldriver::parseHeader (FILE *source, long end, bool *latin1)
{
  long headerlines = 0;
  int name, n_len, addr, a_len;

  delete[] from;
  from = NULL;
  delete[] faddr;
  faddr = NULL;
  delete[] to;
  to = NULL;
  delete[] subject;
  subject = NULL;
  delete[] date;
  date = NULL;

  while (ftell(source) < end)
  {
    fgetsnl(line, MAXLINELEN, source);
    mkstr(line);
    if (!*line) break;

    headerlines++;

    if (strncasecmp(line, "From: ", 6) == 0)
    {
      splitAddress(line + 6, name, n_len, addr, a_len);

      if (n_len)
      {
        from = new char[n_len + 1];
        strncpy(from, line + 6 + name, n_len);
        from[n_len] = '\0';
      }

      if (a_len)
      {
        faddr = new char[a_len + 1];
        strncpy(faddr, line + 6 + addr, a_len);
        faddr[a_len] = '\0';
      }
    }

    else if (strncasecmp(line, "To: ", 4) == 0)
    {
      splitAddress(line + 4, name, n_len, addr, a_len);

      int len = (n_len ? n_len : a_len);
      int pos = (n_len ? name : addr);

      if (len)
      {
        to = new char[len + 1];
        strncpy(to, line + 4 + pos, len);
        to[len] = '\0';
      }
    }

    else if (strncasecmp(line, "Subject: ", 9) == 0)
      subject = strdupplus(line + 9);

    else if (strncasecmp(line, "Date: ", 6) == 0)
      date = strdupplus(line + 6);

    else mimeCharset(line, latin1);
  }

  return headerlines;
}


// return last parseHeader() result
char *alldriver::getFrom () const
{
  return (from ? from : (char *) "");
}


// return last parseHeader() result
char *alldriver::getFaddr () const
{
  return faddr;
}


// return last parseHeader() result
char *alldriver::getTo () const
{
  return (to ? to : (char *) "");
}


// return last parseHeader() result
char *alldriver::getSubject () const
{
  return (subject ? subject : (char *) "");
}


// return last parseHeader() result
char *alldriver::getDate () const
{
  return (date ? date : (char *) "");
}
