/*
 * blueMail offline mail reader
 * Hippo v2 driver

 Copyright (c) 2002 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */


#include <ctype.h>
#include <string.h>
#include "hippo.h"
#include "../../common/auxil.h"
#include "../../common/error.h"
#include "../../common/mysystem.h"


/*
   Hippo v2 main driver
*/

hippo::hippo (bmail *bm)
{
  this->bm = bm;
  alldrv = new alldriver(bm, this);

  NAID = 0;   // next area ID
              // counts the areas already returned by getNextArea()
  NLID = 0;   // next letter ID
              // counts the letters already returned by getNextLetter()
  currentArea = 0;
  maxAreas = 0;
  persmsgs = 0;

  areas = NULL;
  persIDX = NULL;

  persArea = bm->resourceObject->isYes(PersonalArea);
  if (bm->getServiceType() == ST_REPLY) persArea = false;

  body = NULL;
  letterBody = NULL;
  hd = NULL;
}


hippo::~hippo ()
{
  while (maxAreas--)
  {
    delete[] body[maxAreas];
    delete[] areas[maxAreas]->title;
    delete areas[maxAreas];
  }

  delete[] letterBody;
  delete[] body;
  delete[] persIDX;
  delete[] areas;
  delete alldrv;

  if (hd) fclose(hd);
}


const char *hippo::init ()
{
  const char *username = bm->resourceObject->get(UserName);
  bm->resourceObject->set(LoginName, username);
  bm->resourceObject->set(AliasName, username);
  bm->resourceObject->set(SysOpName, "");
  bm->resourceObject->set(BBSName, "");

  const char *hdfile = bm->fileList->exists(".hd");

  if (!(hd = bm->fileList->ftryopen(hdfile, "rt")))
    return "Could not open data file.";

  int len = strrchr(hdfile, '.') - hdfile;
  len = (len > 8 ? 8 : len);
  strncpy(packetID, hdfile, len);
  packetID[len] = '\0';

  persmsgs = buildIndices();

  return NULL;
}


inline int hippo::getNoOfAreas ()
{
  return maxAreas;
}


area_header *hippo::getNextArea ()
{
  area_header *ah;

  bool isPersArea = (persArea && (NAID == 0));

  ah = new area_header(bm, NAID + bm->driverList->getOffset(this),
                       (isPersArea ? "PERS" : ""),
                       (isPersArea ? "PERSONAL" : areas[NAID]->title),
                       (isPersArea ? "Letters addressed to you"
                                   : areas[NAID]->title),
                       (isPersArea ? "Hippo Personal" : "Hippo"),
                       (isPersArea ? A_COLLECTION | A_READONLY |
                                     A_SUBSCRIBED | A_FORCED
                                   : areas[NAID]->flags),
                       areas[NAID]->totmsgs, areas[NAID]->numpers, 71, 71);
  NAID++;

  return ah;
}


void hippo::selectArea (int area)
{
  currentArea = area;
  resetLetters();
}


inline int hippo::getNoOfLetters ()
{
  return areas[currentArea]->totmsgs;
}


letter_header *hippo::getNextLetter ()
{
  char *from = NULL, *to = NULL, *subject = NULL;
  char date[20];
  int flags = 0, msgnum = 0, replyto = 0;

  *date = '\0';

  int area, letter;

  if (persArea && (currentArea == 0))
  {
    area = persIDX[NLID].area;
    letter = persIDX[NLID].msgnum;
  }
  else
  {
    area = currentArea;
    letter = NLID;
  }

  fseek(hd, body[area][letter].pointer, SEEK_SET);
  body[area][letter].msglen = 0;

  while (fgetsnl(line, MAXLINELEN, hd))
  {
    mkstr(line);

    if (strncasecmp(line, "Lines: ", 7) == 0)
    {
      long lines = atol(line + 7);
      for (long i = 0; i < lines; i++)
        body[area][letter].msglen += fgetsnl(line, MAXLINELEN, hd);
      break;
    }

    else if (strncasecmp(line, "Status: ", 8) == 0)
    {
      flags |= (stricstr(line + 8, "Private") ? PRIVATE : 0);
      flags |= (stricstr(line + 8, "Read") ? RECEIVED : 0);
    }

    else if (strncasecmp(line, "Number: ", 8) == 0) msgnum = atoi(line + 8);

    else if (strncasecmp(line, "Reply: ", 7) == 0) replyto = atoi(line + 7);

    else if (strncasecmp(line, "From: ", 6) == 0) from = strdupplus(line + 6);

    else if (strncasecmp(line, "To: ", 4) == 0) to = strdupplus(line + 4);

    else if (strncasecmp(line, "Subject: ", 9) == 0)
      subject = strdupplus(line + 9);

    else if (strncasecmp(line, "Date: ", 6) == 0)
      sprintf(date, "%2.2s.%2.2s.%4.4s %2.2s:%2.2s:%2.2s", line + 12,
                    line + 10, line + 6, line + 14, line + 16, line + 18);
  }

  net_address na;
  if ((areas[area]->flags & A_NETMAIL) && is_digit(*subject)) na.set(subject);

  letter_header *lh = new letter_header(bm, letter,
                                        (from ? from : ""), (to ? to : ""),
                                        (subject ? subject : ""), date,
                                        msgnum, replyto, area, flags,
                                        isLatin1(), &na, this);

  delete[] from;
  delete[] to;
  delete[] subject;

  NLID++;
  return lh;
}


const char *hippo::getBody (int area, int letter, letter_header *lh)
{
  (void) lh;

  int kar;

  delete[] letterBody;
  letterBody = new char[body[area][letter].msglen + 1];
  unsigned char *p = (unsigned char *) letterBody;

  fseek(hd, body[area][letter].pointer, SEEK_SET);
  // skip header
  while (fgetsnl(line, MAXLINELEN, hd))
    if (strncasecmp(line, "Lines: ", 7) == 0) break;

  for (long c = 0; c < body[area][letter].msglen; c++)
  {
    kar = fgetc(hd);

    // a message shouldn't (must not) contain a 0x00 character
    if (kar == 0) kar = ' ';

    if (kar != '\r') *p++ = kar;
  }

  do p--;
  // strip blank lines
  while (p >= (unsigned char *) letterBody && (*p == ' ' || *p == '\n'));

  *++p = '\0';

  return letterBody;
}


inline void hippo::resetLetters ()
{
  NLID = 0;
}


void hippo::initMSF (int **msgstat)
{
  int stat, p = 0;

  // only for real areas, not for collections
  for (int a = persOffset(); a < maxAreas; a++)
  {
    for (int l = 0; l < areas[a]->totmsgs; l++)
    {
      if (p < persmsgs && persIDX[p].area == a && persIDX[p].msgnum == l)
      {
        stat = MSF_PERSONAL;
        p++;
      }
      else stat = 0;

      msgstat[a][l] = stat;
    }
  }
}


int hippo::readMSF (int **msgstat)
{
  alldrv->XTIread(msgstat, persOffset());
  return (MSF_READ | MSF_REPLIED | MSF_PERSONAL | MSF_MARKED);
}


const char *hippo::saveMSF (int **msgstat)
{
  return alldrv->XTIsave(msgstat, persOffset(), packetID);
}


inline bool hippo::isLatin1 ()
{
  // this is either completely right or wrong
  return isIso(bm->resourceObject->get(ConsoleCharset));
}


inline const char **hippo::getBulletins ()
{
  return NULL;
}


inline const char *hippo::getPacketID () const
{
  return packetID;
}


int hippo::persOffset () const
{
  return (persArea ? 1 : 0);
}


int hippo::buildIndices ()
{
  int pers = 0, aix = 0;
  HIPPO_AREA_INFO anchor, *idx = &anchor, *ai = NULL;
  MSG_NDX *mAnchor = new MSG_NDX, *ndx = mAnchor, *i, *next;
  pIDX pAnchor, *pidx = &pAnchor, *d;
  enum {nothing, Message, Info} in = nothing;
  long cmdpos = 0;

  anchor.next = NULL;

  const char *username = bm->resourceObject->get(LoginName);

  if (persArea) idx = newArea(idx);

  while (fgetsnl(line, MAXLINELEN, hd))
  {
    mkstr(line);

    if (strncasecmp(line, "Command: ", 9) == 0)
    {
      cmdpos = ftell(hd);

      if (strcasecmp(line + 9, "Message") == 0) in = Message;
      else if (strcasecmp(line + 9, "Info") == 0) in = Info;
      // no support for NewFiles, Bulletin, File
      else in = nothing;
    }

    else if (strncasecmp(line, "Lines: ", 7) == 0)
    {
      long lines = atol(line + 7);
      for (long j = 0; j < lines; j++) fgetsnl(line, MAXLINELEN, hd);
      in = nothing;
    }

    else switch (in)
    {
      case Message:
        if (strncasecmp(line, "Area: ", 6) == 0)
        {
          aix = getAreaIndex(&anchor, line + 6);

          if (aix == -1)
          {
            idx = newArea(idx);
            idx->title = strdupplus(line + 6);
            aix = maxAreas - 1;
          }

          ndx->area = aix;
          ndx->pointer = cmdpos;

          ai = getArea(&anchor, aix);
          ai->totmsgs++;

          ndx->next = new MSG_NDX;
          ndx = ndx->next;
        }
        else if (strncasecmp(line, "To: ", 4) == 0)
        {
          if (strcasecmp(line + 4, username) == 0)
              /* ||                AliasName */
          {
            pers++;
            ai->numpers++;
            pidx->next = new pIDX;
            pidx = pidx->next;
            pidx->persidx.area = aix;
            pidx->persidx.msgnum = ai->totmsgs - 1;
            pidx->next = NULL;
          }
        }
        break;

      case Info:
        if (strncasecmp(line, "System: ", 8) == 0)
          bm->resourceObject->set(BBSName, line + 8);
        else if (strncasecmp(line, "SysOp: ", 7) == 0)
          bm->resourceObject->set(SysOpName, line + 7);
        else if (strncasecmp(line, "Area: ", 6) == 0)
        {
          aix = getAreaIndex(&anchor, line + 6);

          if (aix == -1)
          {
            idx = newArea(idx);
            idx->title = strdupplus(line + 6);
            aix = maxAreas - 1;
          }

          int flags = 0;
          size_t readlen;

          while ((readlen = fgetsnl(line, MAXLINELEN, hd)))
          {
            if (strncasecmp(line, "Status: ", 8) == 0)
            {
              flags |= (stricstr(line + 8, "Must") ? A_FORCED : 0);
              flags |= (stricstr(line + 8, "Net") ? A_NETMAIL : 0);
              flags |= (stricstr(line + 8, "Mail") ||
                        stricstr(line + 8, "Private") ? A_ALLOWPRIV : 0);
            }
            else if (strncasecmp(line, "Access: ", 8) == 0)
            {
              flags |= (stricstr(line + 8, "Member") ? A_SUBSCRIBED : 0);
              flags |= (stricstr(line + 8, "Write") ? 0 : A_READONLY);
            }
            else
            {
              fseek(hd, ftell(hd) - readlen, SEEK_SET);
              break;
            }
          }

          ai = getArea(&anchor, aix);
          ai->flags = flags;
        }
        break;

      default:
        ;
    }
  }

  if (maxAreas)
  {
    areas = new HIPPO_AREA_INFO *[maxAreas];
    idx = anchor.next;
    int j = 0;

    while (idx)
    {
      areas[j++] = idx;
      idx = idx->next;
    }
  }

  body = new bodytype *[maxAreas];

  for (int a = 0; a < maxAreas; a++)
  {
    body[a] = new bodytype[areas[a]->totmsgs];
    areas[a]->totmsgs = 0;
  }

  i = mAnchor;

  while (i != ndx)
  {
    next = i->next;

    int a = i->area;
    body[a][areas[a]->totmsgs++].pointer = i->pointer;

    delete i;
    i = next;
  }

  delete ndx;

  if (pers)
  {
    persIDX = new PERSIDX[pers];
    pidx = pAnchor.next;
    int i = 0;

    while (pidx)
    {
      persIDX[i].area = pidx->persidx.area;
      persIDX[i].msgnum = pidx->persidx.msgnum;

      i++;
      d = pidx;
      pidx = pidx->next;
      delete d;
    }
  }

  if (persArea) areas[0]->totmsgs = areas[0]->numpers = pers;

  return pers;
}


hippo::HIPPO_AREA_INFO *hippo::newArea (HIPPO_AREA_INFO *idx)
{
  maxAreas++;

  idx->next = new HIPPO_AREA_INFO;
  idx = idx->next;
  idx->title = NULL;
  idx->flags = 0;
  idx->totmsgs = 0;
  idx->numpers = 0;
  idx->next = NULL;

  return idx;
}


int hippo::getAreaIndex (HIPPO_AREA_INFO *anchor, const char *title) const
{
  HIPPO_AREA_INFO *ai = anchor->next;

  // NULL is an internal pseudo-title for the personal area
  if (title == NULL) return 0;

  int a = 0;

  while (ai)
  {
    if (ai->title && (strcasecmp(ai->title, title) == 0)) break;
    ai = ai -> next;
    a++;
  }

  return (ai ? a : -1);
}


hippo::HIPPO_AREA_INFO *hippo::getArea (HIPPO_AREA_INFO *anchor, int num) const
{
  HIPPO_AREA_INFO *ai = anchor->next;

  while (ai)
  {
    if (num-- == 0) break;
    ai = ai -> next;
  }

  return ai;
}


const char *hippo::getAreaInfo (int *area) const
{
  int a = *area;

  *area = areas[a]->flags;
  return areas[a]->title;
}


/*
   Hippo v2 reply driver
*/

//void hippo_reply::replyInf ()
//{
//  FILE *hd;
//  char fname[MYMAXPATH];
//  const char *hdfile = bm->fileList->exists(".hd");
//
//  mkfname(fname, bm->resourceObject->get(InfWorkDir), hdfile);
//
//  if ((hd = fopen(fname, "wt")))
//  {
//    fprintf(hd, "Command: Info\n");
//    fprintf(hd, "System: %s\n", bm->resourceObject->get(BBSName));
//    fprintf(hd, "SysOp: %s\n", bm->resourceObject->get(SysOpName));
//
//    for (int a = mainDriver->persOffset(); a < mainDriver->getNoOfAreas(); a++)
//    {
//      int af = a;
//
//      fprintf(hd, "Area: %s\n", mainDriver->getAreaInfo(&af));
//
//      if (af & (A_FORCED | A_NETMAIL | A_ALLOWPRIV))
//      {
//        int net;
//        bool separate = false;
//
//        fprintf(hd, "Status: ");
//
//        if (af & A_FORCED)
//        {
//          fprintf(hd, "Must");
//          separate = true;
//        }
//
//        if ((net = af & A_NETMAIL))
//        {
//          fprintf(hd, "%sNet", (separate ? "," : ""));
//          separate = true;
//        }
//
//        if (af & A_ALLOWPRIV) fprintf(hd, (net ? "%sMail" : "%sPrivate"),
//                                          (separate ? "," : ""));
//
//        fprintf(hd, "\n");
//      }
//
//      fprintf(hd, "Access: Read");
//      if ((af & A_READONLY) == 0) fprintf(hd, ",Write");
//      if (af & A_SUBSCRIBED) fprintf(hd, ",Member");
//      fprintf(hd, "\n");
//    }
//
//    fprintf(hd, "Lines: 0\n");
//    fclose(hd);
//
//    bm->service->createSystemInfFile(fmtime(hdfile), replyPacketName);
//  }
//}
