/*
 * blueMail offline mail reader
 * Hudson Message Base driver

 Copyright (c) 2002 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */


#include <string.h>
#include "hmb.h"
#include "../../common/auxil.h"
#include "../../common/error.h"
#include "../../common/mysystem.h"


/*
   Hudson Message Base main driver
*/

hmb::hmb (bmail *bm)
{
  this->bm = bm;
  alldrv = new alldriver(bm, this);

  NAID = 0;   // next area ID
              // counts the areas already returned by getNextArea()
  NLID = 0;   // next letter ID
              // counts the letters already returned by getNextLetter()
  currentArea = 0;
  maxAreas = 0;
  persmsgs = 0;

  persHIDX = NULL;

  persArea = bm->resourceObject->isYes(PersonalArea);
  if (bm->getServiceType() == ST_REPLY) persArea = false;

  body = NULL;
  letterBody = NULL;
  idx = hdr = txt = NULL;
}


hmb::~hmb ()
{
  while (maxAreas--)
  {
    delete[] body[maxAreas];
    delete[] board[maxAreas].title;
  }

  delete[] letterBody;
  delete[] body;
  delete[] persHIDX;
  delete alldrv;

  if (idx) fclose(idx);
  if (hdr) fclose(hdr);
  if (txt) fclose(txt);
}


const char *hmb::init ()
{
  FILE *info;

  const char *username = bm->resourceObject->get(UserName);
  bm->resourceObject->set(LoginName, username);
  bm->resourceObject->set(AliasName, username);
  bm->resourceObject->set(SysOpName, "You");
  bm->resourceObject->set(BBSName, "Hudson Message Base");

  for (int i = 0; i < 200; i++) isNetmail[i] = false;

  if (!(info = openFile("info"))) return error;

  size_t rec = fread(&hmb_info, sizeof(hmb_info), 1, info);
  fclose(info);

  if (rec != 1) return "Could not read info file.";

  persmsgs = getPersIndices();          // detects netmail areas, too
  if (persmsgs == -1) return error;

  maxAreas = getAreas();
  body = new bodytype *[maxAreas];
  for (int i = 0; i < maxAreas; i++) body[i] = NULL;

  if (!(txt = openFile("txt")) || !(idx = openFile("idx"))) return error;

  return NULL;
}


inline int hmb::getNoOfAreas ()
{
  return maxAreas;
}


area_header *hmb::getNextArea ()
{
  area_header *ah;
  int totmsgs, numpers, flags;

  bool isPersArea = (persArea && (NAID == 0));

  if (isPersArea) totmsgs = numpers = persmsgs;
  else
  {
    totmsgs = getLSBshort(hmb_info.msgs_on_board[board[NAID].number]);
    numpers = board[NAID].numpers;
  }

  body[NAID] = new bodytype[totmsgs];

  flags = A_SUBSCRIBED | A_FORCED;

  if (!isPersArea && isNetmail[board[NAID].number]) flags |= A_NETMAIL;

  ah = new area_header(bm, NAID + bm->driverList->getOffset(this),
                       (isPersArea ? "PERS" : board[NAID].shortname),
                       (isPersArea ? "PERSONAL" : board[NAID].shortname),
                       (isPersArea ? "Letters addressed to you"
                                   : board[NAID].title),
                       (isPersArea ? "Hudson Personal" : "Hudson"),
                       flags | (isPersArea ? A_COLLECTION | A_READONLY : 0),
                       totmsgs, numpers, 35, 72);
  NAID++;

  return ah;
}


void hmb::selectArea (int area)
{
  currentArea = area;
  resetLetters();
}


int hmb::getNoOfLetters ()
{
  if (persArea && (currentArea == 0)) return persmsgs;
  else return getLSBshort(hmb_info.msgs_on_board[board[currentArea].number]);
}


letter_header *hmb::getNextLetter ()
{
  int area, letter;
  HMB_HDR hmb_hdr;
  int len, flags;
  net_address na;
  unsigned int zone;

  bool isPersArea = (persArea && (currentArea == 0));

  if (isPersArea)
  {
    area = persHIDX[NLID].area;
    letter = persHIDX[NLID].letter;
  }
  else
  {
    area = currentArea;
    letter = NLID;
  }

  getNextHdr(&hmb_hdr, area, isPersArea);

  len = hmb_hdr.from_length;
  char *from = new char[len + 1];
  strncpy(from, hmb_hdr.from, len);
  from[len] = '\0';

  // due to Y2K problems, to_length may be destroyed by date
  len = (hmb_hdr.date_length == 9 ? sizeof(hmb_hdr.to) : hmb_hdr.to_length);
  char *to = new char[len + 1];
  strncpy(to, hmb_hdr.to, len);
  to[len] = '\0';

  len = hmb_hdr.subj_length;
  char *subject = new char[len + 1];
  strncpy(subject, hmb_hdr.subject, len);
  subject[len] = '\0';

  len = hmb_hdr.date_length + 2 + hmb_hdr.time_length;
  char *date = new char[len + 1];
  strncpy(date, hmb_hdr.date, hmb_hdr.date_length);
  date[hmb_hdr.date_length] = '\0';
  // due to Y2K problems, the date may be too long
  if (hmb_hdr.date_length == 9)
  {
    date[6] = date[7];
    date[7] = date[8];
    date[8] = date[9];
  }
  strcat(date, "  ");
  strncat(date, hmb_hdr.time, hmb_hdr.time_length);

  flags = (hmb_hdr.msg_attr & HMB_MSG_PRIVATE ? PRIVATE : 0) |
          (hmb_hdr.msg_attr & HMB_MSG_RECEIVED ? RECEIVED : 0) |
          (hmb_hdr.msg_attr & HMB_MSG_OUTECHO ? INTRANSIT : 0) |   // ?
          (hmb_hdr.msg_attr & HMB_MSG_LOCAL ? LOCAL : 0) |
          (hmb_hdr.net_attr & HMB_NET_KILL ? KILL : 0) |
          (hmb_hdr.net_attr & HMB_NET_SENT ? SENT : 0) |
          (hmb_hdr.net_attr & HMB_NET_FATTACH ? FILEATTACHED : 0) |
          (hmb_hdr.net_attr & HMB_NET_CRASH ? CRASH : 0);

  size_t size = sizeof(HMB_TXT);
  body[area][letter].pointer = getLSBshort(hmb_hdr.start_record) * size;
  size -= sizeof(unsigned char);
  body[area][letter].msglen = getLSBshort(hmb_hdr.records) * size;

  if ((zone = hmb_hdr.orig_zone)) na.set(zone, getLSBshort(hmb_hdr.orig_net),
                                         getLSBshort(hmb_hdr.orig_node), 0);
  // point only available through message (body)

  letter_header *lh = new letter_header(bm, letter, from, to, subject, date,
                                        getLSBshort(hmb_hdr.msgnum),
                                        getLSBshort(hmb_hdr.prev_reply),
                                        area, flags, isLatin1(), &na, this);

  delete[] from;
  delete[] to;
  delete[] subject;
  delete[] date;

  NLID++;
  return lh;
}


const char *hmb::getBody (int area, int letter, letter_header *lh)
{
  HMB_TXT hmb_txt;
  int kar;

  delete[] letterBody;
  letterBody = new char[body[area][letter].msglen + 1];
  unsigned char *p = (unsigned char *) letterBody;

  fseek(txt, body[area][letter].pointer, SEEK_SET);
  unsigned int chunks = body[area][letter].msglen / (sizeof(HMB_TXT) -
                                                     sizeof(unsigned char));

  while (chunks--)
  {
    if (fread(&hmb_txt, sizeof(HMB_TXT), 1, txt) != 1)
      fatalError("Could not get next letter.");

    for (int c = 0; c < hmb_txt.length; c++)
    {
      kar = hmb_txt.chunk[c];

      // a message shouldn't (must not) contain a 0x00 character
      if (kar == 0) kar = ' ';

      // end of line?
      if (kar == '\r') *p++ = '\n';
      else if (kar != '\n') *p++ = kar;
    }
  }

  do p--;
  // strip blank lines
  while (p >= (unsigned char *) letterBody && (*p == ' ' || *p == '\n'));

  *++p = '\0';

  // determine charset
  bool latin1 = lh->isLatin1();
  alldrv->fsc0054(letterBody, '\1', &latin1);
  lh->setLatin1(latin1);

  // get netmail address from echomail
  net_address &na = lh->getNetAddr();
  char *addr = alldrv->getFidoAddr(letterBody);
  if (addr) na.set(addr);

  // add point to netmail address
  unsigned int point;
  char *fmpt = strstr(letterBody, "\1FMPT ");
  if (fmpt && sscanf(fmpt, "\1FMPT %u\n", &point))
    na.set(na.getZone(), na.getNet(), na.getNode(), point);

  // decode quoted-printable
  if (bm->driverList->allowQuotedPrintable() && alldrv->isQP(letterBody))
    mimeDecodeBody(letterBody);

  return letterBody;
}


void hmb::resetLetters ()
{
  NLID = 0;
  fseek(hdr, 0, SEEK_SET);
}


inline void hmb::initMSF (int **msgstat)
{
  (void) msgstat;
}


int hmb::readMSF (int **msgstat)
{
  HMB_IDX hmb_idx;
  unsigned int msg;
  unsigned char bidx;
  int a, stat, p = 0;

  unsigned int *l = new unsigned int[maxAreas];
  for (a = 0; a < maxAreas; a++) l[a] = 0;

  while (fread(&hmb_idx, sizeof(HMB_IDX), 1, idx) == 1)
  {
    msg = getLSBshort(hmb_idx.msgnum);

    // if not marked for deletion
    if (msg != 0xFFFF)
    {
      bidx = hmb_idx.board - 1;
      a = getAreaIndex(bidx);

      if (a != maxAreas)
        if (l[a] < getLSBshort(hmb_info.msgs_on_board[bidx]))
        {
          stat = (msg <= board[a].lastread ? MSF_READ : 0);
          if (p < persmsgs && persHIDX[p].msgnum == msg)
          {
            board[a].numpers++;   // count personal messages
            persHIDX[p].area = a;
            persHIDX[p].letter = l[a];
            p++;
            stat |= MSF_PERSONAL;
          }
          msgstat[a][l[a]++] = stat;
        }
    }
  }

  delete[] l;
  return (MSF_READ | MSF_PERSONAL);
}


const char *hmb::saveMSF (int **msgstat)
{
  FILE *file;
  UINT lastread;
  HMB_HDR hmb_hdr;

  memset(&lastread, 0, sizeof(lastread));

  if ((file = bm->fileList->ftryopen("lastread.bbs", "Xr+b")))
  {
    for (int i = 0; i < 200; i++)
      if (fwrite(&lastread, sizeof(lastread), 1, file) != 1)
      {
        fclose(file);
        return NULL;
      }

    for (int a = (persArea ? 1 : 0); a < maxAreas; a++)
    {
      selectArea(a);
      memset(&lastread, 0, sizeof(lastread));

      for (int l = 0; l < getNoOfLetters(); l++)
      {
        getNextHdr(&hmb_hdr, a, false);
        if (msgstat[a][l] & MSF_READ)
          memcpy(&lastread, hmb_hdr.msgnum, sizeof(lastread));
      }

      if (fseek(file, board[a].number << 1, SEEK_SET) != 0 ||
          fwrite(&lastread, sizeof(lastread), 1, file) != 1)
      {
        fclose(file);
        return NULL;
      }
    }

    fclose(file);
    return "";
  }
  else return NULL;
}


inline bool hmb::isLatin1 ()
{
  return false;
}


inline const char **hmb::getBulletins ()
{
  return NULL;
}


FILE *hmb::openFile (const char *which)
{
  FILE *file;
  char buf[13];

  sprintf(buf, "msg%s.bbs", which);

  if (!(file = bm->fileList->ftryopen(buf, "xrb")))
    sprintf(error, "Could not lock %s file.", which);

  return file;
}


int hmb::getPersIndices ()
{
  int pers = 0;
  HMB_HDR hmb_hdr;
  struct hIDX
  {
    unsigned int msgnum;
    hIDX *next;
  } *anchor = new hIDX, *pidx = anchor, *i, *next;

  if (!(hdr = openFile("hdr")))
  {
    delete pidx;
    return -1;
  }

  const char *username = bm->resourceObject->get(LoginName);
  int ulen = strlen(username);

  while (fread(&hmb_hdr, sizeof(HMB_HDR), 1, hdr) == 1)
    if ((hmb_hdr.msg_attr & HMB_MSG_DELETED) == 0)
    {
      // detect netmail areas
      if (hmb_hdr.msg_attr & HMB_MSG_NETMAIL)
        isNetmail[hmb_hdr.board - 1] = true;

      // detect personal messages
      if (ulen == hmb_hdr.to_length &&
          strncasecmp(hmb_hdr.to, username, hmb_hdr.to_length) == 0)
          /* || (alen &&          AliasName) */
      {
        pidx->msgnum = getLSBshort(hmb_hdr.msgnum);
        pidx->next = new hIDX;
        pidx = pidx->next;
        pers++;
      }
    }

  persHIDX = new PERSHIDX[pers];
  pers = 0;
  i = anchor;

  while (i != pidx)
  {
    next = i->next;
    persHIDX[pers++].msgnum = i->msgnum;
    delete i;
    i = next;
  }

  delete pidx;

  return pers;
}


int hmb::getAreas ()
{
  FILE *file;
  char buffer[54], bbsname[50];
  int areas = (persArea ? 1 : 0), items = 0, bbsnumber;
  static char unavailable[] = "unavailable";
  UINT lastread;

  if (persArea)
  {
    board[0].number = 255;   // internal pseudo-number for the personal area
    board[0].title = NULL;
  }

  const char *areasBBS =
    (bm->getServiceType() == ST_REPLY ? "areas.bbs"
                                      : bm->resourceObject->get(AreasBBS));

  if (areasBBS && (file = fopen(areasBBS, "rt")))
  {
    while (fgets(buffer, sizeof(buffer), file))
    {
      if (*buffer != ';' && *buffer != '\n')
        if ((items = sscanf(buffer, "%d %49s", &bbsnumber, bbsname)) &&
            bbsnumber >= 1 && bbsnumber <= 200)
        {
          board[areas].number = bbsnumber - 1;
          sprintf(board[areas].shortname, "%d", bbsnumber);
          board[areas].title = strdupplus(items == 2 ? bbsname : unavailable);
          board[areas].lastread = 0;
          board[areas].numpers = 0;
          areas++;
        }
    }
    fclose(file);
    if (areas == 0) fatalError("Could not read any area.");
  }
  else
  {
    for (int i = 0, j = (persArea ? 1 : 0); i < 200; i++, j++)
    {
      board[j].number = i;
      sprintf(board[j].shortname, "%d", i + 1);
      board[j].title = strdupplus(unavailable);
      board[j].lastread = 0;
      board[j].numpers = 0;
      areas++;
    }
  }

  if ((file = bm->fileList->ftryopen("lastread.bbs", "rb")))
  {
    for (int i = (persArea ? 1 : 0); i < areas; i++)
      if (fseek(file, board[i].number << 1, SEEK_SET) == 0)
        if (fread(&lastread, sizeof(lastread), 1, file) == 1)
          board[i].lastread = getLSBshort(lastread);
    fclose(file);
  }

  return areas;
}


void hmb::getNextHdr (HMB_HDR *hmb_hdr, int area, bool isPersArea) const
{
  bool fok;

  while ((fok = (fread(hmb_hdr, sizeof(HMB_HDR), 1, hdr) == 1)))
    if ((hmb_hdr->msg_attr & HMB_MSG_DELETED) == 0)
    {
      if (isPersArea)
      {
        if (getLSBshort(hmb_hdr->msgnum) == persHIDX[NLID].msgnum) break;
      }
      else
      {
        if (hmb_hdr->board == board[area].number + 1) break;
      }
    }

  if (!fok) fatalError("Header file corrupt.");
}


int hmb::getAreaIndex (unsigned char area) const
{
  int a;

  // 255 is an internal pseudo-number for the personal area
  if (area == 255) return 0;

  for (a = (persArea ? 1 : 0); a < maxAreas; a++)
    if (board[a].number == area) break;

  return a;
}


/*
   Hudson Message Base reply driver
*/

//void hmb_reply::replyInf ()
//{
//  HMB_INFO hmb_info;
//  char fname[MYMAXPATH];
//  const char *dir = bm->resourceObject->get(InfWorkDir);
//  FILE *file;
//
//  memset(&hmb_info, 0, sizeof(hmb_info));
//  mkfname(fname, dir, "msginfo.bbs");
//
//  if ((file = fopen(fname, "wb")))
//  {
//    fwrite(&hmb_info, sizeof(hmb_info), 1, file);
//    fclose(file);
//  }
//
//  fcreate(dir, "msgidx.bbs");
//  fcreate(dir, "msghdr.bbs");
//  fcreate(dir, "msgtxt.bbs");
//
//  const char *areasBBS = bm->resourceObject->get(AreasBBS);
//
//  if (areasBBS)
//  {
//    mkfname(fname, dir, "areas.bbs");
//    fcopy(areasBBS, fname);
//  }
//
//  bm->service->createSystemInfFile((areasBBS ? fmtime(areasBBS) : 0),
//                                   replyPacketName);
//}
