/*
 * blueMail offline mail reader
 * OMEN driver

 Copyright (c) 2002 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */


#include <string.h>
#include "omen.h"
#include "../../common/auxil.h"
#include "../../common/error.h"
#include "../../common/mysystem.h"


/*
   OMEN main driver
*/

omen::omen (bmail *bm)
{
  this->bm = bm;
  alldrv = new alldriver(bm, this);

  NAID = 0;   // next area ID
              // counts the areas already returned by getNextArea()
  NLID = 0;   // next letter ID
              // counts the letters already returned by getNextLetter()
  currentArea = 0;
  maxAreas = 0;
  persmsgs = 0;

  areas = NULL;
  persIDX = NULL;

  persArea = bm->resourceObject->isYes(PersonalArea);
  if (bm->getServiceType() == ST_REPLY) persArea = false;

  c_set_iso = false;

  body = NULL;
  letterBody = NULL;
  bulletins = NULL;
  newmsg = NULL;
}


omen::~omen ()
{
  while (maxAreas--)
  {
    delete[] body[maxAreas];
    delete[] areas[maxAreas].title;
  }

  delete[] bulletins;
  delete[] letterBody;
  delete[] body;
  delete[] persIDX;
  delete[] areas;
  delete alldrv;

  if (newmsg) fclose(newmsg);
}


const char *omen::init ()
{
  FILE *system;

  const char *username = bm->resourceObject->get(UserName);
  bm->resourceObject->set(LoginName, username);
  bm->resourceObject->set(AliasName, username);

  const char *sysfile = bm->fileList->exists("system*");

  bool areas_ok = ((system = bm->fileList->ftryopen(sysfile, "rb")) &&
                   getAreas(system));

  if (system) fclose(system);
  if (!areas_ok) return "Could not access system file.";

  strncpy(packetID, sysfile + 6, 2);
  packetID[2] = '\0';

  readInfo();

  if ((persmsgs = buildIndices()) == -1) return "Could not read newmsg file.";

  const char bulletinfile[2][13] = {"bullet*", "nfile*"};
  bulletins = alldrv->bulletin_list(bulletinfile, 2, BL_NOADD);

  return NULL;
}


inline int omen::getNoOfAreas ()
{
  return maxAreas;
}


area_header *omen::getNextArea ()
{
  area_header *ah;
  char areanum[12];
  int flags;

  bool isPersArea = (persArea && (NAID == 0));

  sprintf(areanum, "%5d", areas[NAID].number);

  if (isPersArea) flags = A_COLLECTION | A_READONLY | A_SUBSCRIBED | A_FORCED;
  else flags = (areas[NAID].flags & BrdStatus_SELECTED ? A_SUBSCRIBED : 0) |
               (areas[NAID].flags & BrdStatus_NETMAIL ? A_NETMAIL : 0) |
               (areas[NAID].flags & BrdStatus_WRITE ? 0 : A_READONLY) |
               (areas[NAID].flags & BrdStatus_ALIAS ? A_ALIAS : 0) |
               (areas[NAID].flags & BrdStatus_PRIVATE ? A_ALLOWPRIV : 0);

  ah = new area_header(bm, NAID + bm->driverList->getOffset(this),
                       (isPersArea ? "PERS" : areanum),
                       (isPersArea ? "PERSONAL" : areas[NAID].shortname),
                       (isPersArea ? "Letters addressed to you"
                                   : (areas[NAID].title ? areas[NAID].title
                                                        : areas[NAID].shortname)),
                       (isPersArea ? "OMEN Personal" : "OMEN"),
                       flags, areas[NAID].totmsgs, areas[NAID].numpers,
                       35, 72);
  NAID++;

  return ah;
}


void omen::selectArea (int area)
{
  currentArea = area;
  resetLetters();
}


inline int omen::getNoOfLetters ()
{
  return areas[currentArea].totmsgs;
}


letter_header *omen::getNextLetter ()
{
  int area, letter;
  char *p, *q, *date, *faddr, subject[73];
  long prev = 0, next;
  net_address na;

  if (persArea && (currentArea == 0))
  {
    area = persIDX[NLID].area;
    letter = persIDX[NLID].msgnum;
  }
  else
  {
    area = currentArea;
    letter = NLID;
  }

  fseek(newmsg, body[area][letter].pointer, SEEK_SET);
  fgetsnl(line, MAXLINELEN, newmsg);

  p = strchr(line, ' ');
  *p = '\0';
  int msgnum = (int) atol(line + 2);     // we're somewhat limited here

  p = strstr(p + 1, "  ");
  q = strchr(p + 1, '(');
  *(q - 2) = '\0';
  date = strdupplus(p + 2);

  sscanf(q, "(%ld/%ld)", &prev, &next);
  int replyto = (int) prev;            // we're somewhat limited here

  p = strchr(q + 1, '(');

  int flags = (strchr(p, 'P') ? PRIVATE : 0) |
              (strchr(p, 'R') ? RECEIVED : 0);

  fgetsnl(line, MAXLINELEN, newmsg);

  p = strstr(mkstr(line), " => ");
  *p = '\0';
  q = p + 4;

  if ((faddr = stricstr(q, " From: (")))
  {
    na.set(faddr + 8);
    *faddr = '\0';
  }

  // skip "Subj: "
  fread(subject, 1, 6, newmsg);

  int c;
  unsigned int i = 0;

  do
  {
    c = fgetc(newmsg);
    subject[i++] = c;
  }
  while (c != 2 && i < sizeof(subject));

  subject[i - 1] = '\0';

  NLID++;
  letter_header *lh = new letter_header(bm, letter, line, q, subject, date,
                                        msgnum, replyto, area, flags,
                                        isLatin1(), &na, this);
  delete[] date;

  return lh;
}


const char *omen::getBody (int area, int letter, letter_header *lh)
{
  int kar;

  delete[] letterBody;
  letterBody = new char[body[area][letter].msglen + 1];
  unsigned char *p = (unsigned char *) letterBody;

  fseek(newmsg, body[area][letter].pointer, SEEK_SET);
  // skip header
  while (!feof(newmsg) && (fgetc(newmsg) != 2));

  for (long c = 0; c < body[area][letter].msglen; c++)
  {
    kar = fgetc(newmsg);

    // a message shouldn't (must not) contain a 0x00 character
    if (kar == 0) kar = ' ';

    if (kar != '\r') *p++ = kar;
  }

  do p--;
  // strip blank lines
  while (p >= (unsigned char *) letterBody && (*p == ' ' || *p == '\n'));

  *++p = '\0';

  // determine charset
  bool latin1 = lh->isLatin1();
  alldrv->fsc0054(letterBody, '\1', &latin1);
  lh->setLatin1(latin1);

  // get netmail address from echomail
  net_address &na = lh->getNetAddr();
  char *addr = alldrv->getFidoAddr(letterBody);
  if (addr) na.set(addr);

  // add point to netmail address
  unsigned int point;
  char *fmpt = strstr(letterBody, "\1FMPT ");
  if (fmpt && sscanf(fmpt, "\1FMPT %u\n", &point))
    na.set(na.getZone(), na.getNet(), na.getNode(), point);

  // decode quoted-printable
  if (bm->driverList->allowQuotedPrintable() && alldrv->isQP(letterBody))
    mimeDecodeBody(letterBody);

  return letterBody;
}


inline void omen::resetLetters ()
{
  NLID = 0;
}


void omen::initMSF (int **msgstat)
{
  int stat, p = 0;

  // only for real areas, not for collections
  for (int a = persOffset(); a < maxAreas; a++)
  {
    for (int l = 0; l < areas[a].totmsgs; l++)
    {
      if (p < persmsgs && persIDX[p].area == a && persIDX[p].msgnum == l)
      {
        stat = MSF_PERSONAL;
        p++;
      }
      else stat = 0;

      msgstat[a][l] = stat;
    }
  }
}


int omen::readMSF (int **msgstat)
{
  alldrv->XTIread(msgstat, persOffset());
  return (MSF_READ | MSF_REPLIED | MSF_PERSONAL | MSF_MARKED);
}


const char *omen::saveMSF (int **msgstat)
{
  return alldrv->XTIsave(msgstat, persOffset(), packetID);
}


inline bool omen::isLatin1 ()
{
  return c_set_iso;
}


inline const char **omen::getBulletins ()
{
  return bulletins;
}


inline const char *omen::getPacketID () const
{
  return packetID;
}


int omen::persOffset () const
{
  return (persArea ? 1 : 0);
}


bool omen::getAreas (FILE *sysfile)
{
  struct
  {
    unsigned char BrdNum;
    unsigned char BrdStatus;
    unsigned char BrdHighNum;
    char BrdName[17];         // 1 byte length + 16 char (Pascal style string)
  } BrdRecord;

  // system name is 1 byte length + 40 char (Pascal style string)
  if (fread(line, 41, 1, sysfile) != 1) return false;
  line[*line + 1] = '\0';     // make it a C style string

  bm->resourceObject->set(BBSName, line + 1);

  maxAreas = (bm->fileList->getSize() - 41) / sizeof(BrdRecord) + persOffset();

  areas = new OMEN_AREA_INFO[maxAreas];

  if (persArea)
  {
    areas[0].title = NULL;
    areas[0].totmsgs = areas[0].numpers = 0;
  }

  FILE *bnames = bm->fileList->ftryopen("bnames*", "rt");

  int i = persOffset();

  while (fread(&BrdRecord, sizeof(BrdRecord), 1, sysfile) == 1)
  {
    areas[i].number = (BrdRecord.BrdHighNum << 8) + BrdRecord.BrdNum;
    strncpy(areas[i].shortname, BrdRecord.BrdName + 1, *BrdRecord.BrdName);
    areas[i].shortname[*BrdRecord.BrdName] = '\0';
    areas[i].title = NULL;
    areas[i].flags = BrdRecord.BrdStatus;
    areas[i].totmsgs = 0;
    areas[i].numpers = 0;

    if (bnames && fgetsnl(line, MAXLINELEN, bnames))
      areas[i].title = strdupplus(strchr(mkstr(line), ':') + 1);

    i++;
  }

  if (bnames) fclose(bnames);

  return true;
}


void omen::readInfo ()
{
  FILE *info;

  if ((info = bm->fileList->ftryopen("info*", "rt")))
  {
    while (fgetsnl(line, MAXLINELEN, info))
    {
      mkstr(line);

      if (strncasecmp(line, "SYSOP:", 6) == 0)
        bm->resourceObject->set(SysOpName, line + 6);
      else if (strcasecmp(line, "C_SET:ISO") == 0) c_set_iso = true;
    }
    fclose(info);
  }
  else bm->resourceObject->set(SysOpName, "");
}


int omen::buildIndices ()
{
  size_t readlen;
  char MsgNum[11];
  int BoardNum, pers = 0;
  char *separator, *faddr;
  pIDX pAnchor, *pidx = &pAnchor, *d;

  body = new bodytype *[maxAreas];
  for (int i = 0; i < maxAreas; i++) body[i] = NULL;

  if (!(newmsg = bm->fileList->ftryopen("newmsg*", "rb"))) return -1;

  struct MSG_IDX
  {
    int area;
    long pointer;
    long msglen;
    MSG_IDX *next;
  } *mAnchor = new MSG_IDX, *ndx = mAnchor, *i, *next;

  const char *username = bm->resourceObject->get(LoginName);

  while ((readlen = fgetsnl(line, MAXLINELEN, newmsg)))
  {
    if (sscanf(line, "\1#%s  %d:", MsgNum, &BoardNum) == 2)
    {
      int aix = getAreaIndex(BoardNum);

      ndx->area = aix;
      ndx->pointer = ftell(newmsg) - readlen;
      ndx->msglen = 0;
      areas[aix].totmsgs++;

      i = ndx;

      ndx->next = new MSG_IDX;
      ndx = ndx->next;

      if (fgetsnl(line, MAXLINELEN, newmsg) &&
          (separator = strstr(mkstr(line), " => ")))
      {
        if ((faddr = stricstr(separator + 4, " From: ("))) *faddr = '\0';
        if (strcasecmp(separator + 4, username) == 0)
            /* ||               AliasName */
        {
          pers++;
          areas[aix].numpers++;
          pidx->next = new pIDX;
          pidx = pidx->next;
          pidx->persidx.area = aix;
          pidx->persidx.msgnum = areas[aix].totmsgs - 1;
          pidx->next = NULL;
        }
      }

      // skip header
      while (!feof(newmsg) && (fgetc(newmsg) != 2));

      long pos = ftell(newmsg);

      // skip body
      while (!feof(newmsg) && (fgetc(newmsg) != 3));

      i->msglen = ftell(newmsg) - pos - 1;
    }
  }

  for (int a = 0; a < maxAreas; a++)
  {
    body[a] = new bodytype[areas[a].totmsgs];
    areas[a].totmsgs = 0;
  }

  i = mAnchor;

  while (i != ndx)
  {
    next = i->next;

    int a = i->area;
    body[a][areas[a].totmsgs].pointer = i->pointer;
    body[a][areas[a].totmsgs].msglen = i->msglen;
    areas[a].totmsgs++;

    delete i;
    i = next;
  }

  delete ndx;

  if (pers)
  {
    persIDX = new PERSIDX[pers];
    pidx = pAnchor.next;
    int i = 0;

    while (pidx)
    {
      persIDX[i].area = pidx->persidx.area;
      persIDX[i].msgnum = pidx->persidx.msgnum;

      i++;
      d = pidx;
      pidx = pidx->next;
      delete d;
    }
  }

  if (persArea) areas[0].totmsgs = areas[0].numpers = pers;

  return pers;
}


int omen::getAreaIndex (int area) const
{
  int a;

  // -1 is an internal pseudo-number for the personal area
  if (area == -1) return 0;

  for (a = persOffset(); a < maxAreas; a++)
    if (areas[a].number == area) break;

  return (a == maxAreas ? -1 : a);
}


/*
   OMEN reply driver
*/

//void omen_reply::replyInf ()
//{
//  char fname[MYMAXPATH];
//
//  const char *sysfile = bm->fileList->exists("system*");
//  const char *newmsg = bm->fileList->exists("newmsg*");
//  const char *bnames = bm->fileList->exists("bnames*");
//  const char *info = bm->fileList->exists("info*");
//  const char *dir = bm->resourceObject->get(InfWorkDir);
//
//  mkfname(fname, dir, sysfile);
//  fcopy(sysfile, fname);
//
//  fcreate(dir, newmsg);
//
//  if (bnames)
//  {
//    mkfname(fname, dir, bnames);
//    fcopy(bnames, fname);
//  }
//
//  if (info)
//  {
//    mkfname(fname, dir, info);
//    fcopy(info, fname);
//  }
//
//  bm->service->createSystemInfFile(fmtime(sysfile), replyPacketName);
//}
