/*
 * blueMail offline mail reader
 * QWK / QWKE packet driver

 Copyright (c) 1997 John Zero <john@graphisoft.hu>
 Copyright (c) 1998 William McBrine <wmcbrine@clark.net>
 Copyright (c) 2002 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */


#include <stdlib.h>
#include <string.h>
#include "qwk.h"
#include "../../common/auxil.h"
#include "../../common/error.h"
#include "../../common/mysystem.h"


/*
   QWK header (to handle the header in the MESSAGES.DAT file)
*/

void qwk_header::qstrncpy (char *dest, const char *source, size_t maxlen)
{
  char *p;

  strncpy(dest, source, maxlen);
  dest[maxlen] = '\0';
  for (p = strrchr(dest, '\0') - 1; p > dest && *p == ' '; p--);
  *++p = '\0';
}


bool qwk_header::read (FILE *msgFile)
{
  QWK_MSGHEADER qmh;
  char buf[9];

  if (fread(&qmh, sizeof(qmh), 1, msgFile) != 1) return false;

  privat = (qmh.status == '*' || qmh.status == '+');
  skipit = (qmh.status == '\0' || qmh.status == '\xFF');

  if (skipit) return true;

  qstrncpy(from, qmh.from, 25);
  qstrncpy(to, qmh.to, 25);
  qstrncpy(subject, qmh.subject, 25);

  qstrncpy(date, qmh.date, 8);
  //date[2] = '-';
  //date[5] = '-';                // to deal with some broken messages
  strcat(date, " ");
  strncat(date, qmh.time, 5);

  qstrncpy(buf, qmh.msgnum, 7);
  msgnum = atoi(buf);

  qstrncpy(buf, qmh.refnum, 8);
  refnum = atoi(buf);

  confnum = getLSBshort(qmh.confnum);

  qstrncpy(buf, qmh.chunks, 6);
  msglen = (atoi(buf) - 1) << 7;   // 1 record is the header itself

  return true;
}


bool qwk_header::write (FILE *msgFile)
{
  QWK_MSGHEADER qmh;
  char buf[9];
  int len;
  long chunks;

  memset(&qmh, ' ', sizeof(qmh));

  if (privat) qmh.status = '*';

  sprintf(buf, " %-6d", confnum);
  strncpy(qmh.msgnum, buf, 7);     // is confnum in replies

  strncpy(qmh.date, date, 8);
  strncpy(qmh.time, &date[9], 5);

  len = strlen(to);
  strncpy(qmh.to, to, (len < 25 ? len : 25));

  len = strlen(from);
  strncpy(qmh.from, from, (len < 25 ? len : 25));

  len = strlen(subject);
  strncpy(qmh.subject, subject, (len < 25 ? len : 25));

  if (refnum)
  {
    sprintf(buf, " %-7d", refnum);
    strncpy(qmh.refnum, buf, 8);
  }

  chunks = ((msglen + MSGDAT_RECLEN - 1) >> 7) + 1;     // 1 record for header
  sprintf(buf, "%-6ld", chunks);
  strncpy(qmh.chunks, buf, 6);

  qmh.alive = (char) 0xE1;     // message is active

  putLSBshort(qmh.confnum, confnum);

  return (fwrite(&qmh, sizeof(qmh), 1, msgFile) == 1);
}


/*
   QWK / QWKE main driver
*/

qwk::qwk (bmail *bm)
{
  this->bm = bm;
  alldrv = new alldriver(bm, this);

  NAID = 0;   // next area ID
              // counts the areas already returned by getNextArea()
  NLID = 0;   // next letter ID
              // counts the letters already returned by getNextLetter()
  currentArea = 0;
  maxAreas = 0;

  areas = NULL;

  const char *pa = bm->resourceObject->get(PersonalArea);
  bool pndx = (bm->fileList->exists("personal.ndx") != NULL);
  if (!pndx) pndx = (bm->fileList->exists(".ndx") == NULL);
  if (bm->getServiceType() == ST_REPLY) pa = "N";

  persArea = (pa && (toupper(*pa) == 'Y') ? true : (pa ? false : pndx));
  qwke = (bm->fileList->exists("toreader.ext") != NULL);

  body = NULL;
  letterBody = NULL;
  bulletins = NULL;
  msgFile = NULL;
}


qwk::~qwk ()
{
  while (maxAreas--)
  {
    delete[] body[maxAreas];
    delete[] areas[maxAreas].title;
  }

  delete[] bulletins;
  delete[] letterBody;
  delete[] body;
  delete[] areas;
  delete alldrv;

  if (msgFile) fclose(msgFile);
}


const char *qwk::init ()
{
  if (!readControl()) return error;

  body = new bodytype *[maxAreas];
  for (int i = 0; i < maxAreas; i++) body[i] = NULL;

  if (qwke && !readExtInformation()) return error;

  if (!(msgFile = openFile("messages.dat"))) return error;

  if (!readIndices()) return error;

  return NULL;
}


inline int qwk::getNoOfAreas ()
{
  return maxAreas;
}


area_header *qwk::getNextArea ()
{
  area_header *ah;

  bool isPersArea = (persArea && (NAID == 0));

  int maxfromto = (qwke ? QWKE_MAXFROMTO : 25);
  int maxsubject = (qwke ? QWKE_MAXSUBJECT : 25);

  ah = new area_header(bm, NAID + bm->driverList->getOffset(this),
                       (isPersArea ? "PERS" : areas[NAID].areanum),
                       (isPersArea ? "PERSONAL" : areas[NAID].title),
                       (isPersArea ? "Letters addressed to you"
                                   : areas[NAID].title),
                       (isPersArea ? (qwke ? "QWKE Personal" : "QWK Personal")
                                   : (qwke ? "QWKE" :"QWK")),
                       (isPersArea ? A_COLLECTION | A_READONLY |
                                     A_SUBSCRIBED | A_FORCED
                                   : areas[NAID].flags),
                       areas[NAID].totmsgs, 0, maxfromto, maxsubject);
  NAID++;

  return ah;
}


void qwk::selectArea (int area)
{
  currentArea = area;
  resetLetters();
}


inline int qwk::getNoOfLetters ()
{
  return areas[currentArea].totmsgs;
}


letter_header *qwk::getNextLetter ()
{
  int area, letter;
  qwk_header qh;

  fseek(msgFile, body[currentArea][NLID].pointer - MSGDAT_RECLEN, SEEK_SET);
  if (!qh.read(msgFile)) fatalError("Could not get next letter.");

  if (persArea && (currentArea == 0))
  {
    area = getAreaIndex(qh.confnum);
    letter = getLetterIndex(area, body[currentArea][NLID].pointer);
  }
  else
  {
    area = currentArea;
    letter = NLID;
  }

  body[area][letter].msglen = qh.msglen;

  NLID++;
  return new letter_header(bm, letter,
                           qh.from, qh.to, qh.subject, qh.date,
                           qh.msgnum, qh.refnum, area,
                           (qh.privat ? PRIVATE : 0),
                           ((areas[area].flags & A_INTERNET) != 0),
                           NULL, this);
}


const char *qwk::getBody (int area, int letter, letter_header *lh)
{
  int kar;
  bool noSoftCR = bm->resourceObject->isYes(StripSoftCR);

  delete[] letterBody;
  letterBody = new char[body[area][letter].msglen + 1];
  unsigned char *p = (unsigned char *) letterBody;
  fseek(msgFile, body[area][letter].pointer, SEEK_SET);

  for (long c = 0; c < body[area][letter].msglen; c++)
  {
    kar = fgetc(msgFile);

    // a message shouldn't (must not) contain a 0x00 character
    if (kar == 0) kar = ' ';

    // strip soft CR
    if (noSoftCR && (kar == SOFT_CR)) continue;

    // end of line?
    if (kar == 227) *p++ = '\n';
    else if (kar != '\r') *p++ = kar;
  }

  do p--;
  // strip blank lines
  while (p >= (unsigned char *) letterBody && (*p == ' ' || *p == '\n'));

  *++p = '\0';

  // determine charset
  bool latin1 = lh->isLatin1();
  alldrv->fsc0054(letterBody, '\1', &latin1);
  lh->setLatin1(latin1);

  // get netmail address from echomail (for addressbook entry etc. only)
  net_address &na = lh->getNetAddr();
  char *addr = alldrv->getFidoAddr(letterBody);
  if (addr) na.set(addr);

  // decode quoted-printable
  if (bm->driverList->allowQuotedPrintable() && alldrv->isQP(letterBody))
    mimeDecodeBody(letterBody);

  char *body = letterBody;

  if (qwke)
  {
    bool isEmail = false, QWKEKludge;
    char *p;
    char kline[QWKE_MAXSUBJECT + 1];

    // handle netmail

    if (areas[area].flags & A_NETMAIL)
    {
      if (areas[area].flags & A_INTERNET) isEmail = true;
      else if (*lh->getSubject() == '@') na.set(lh->getSubject() + 1);
    }

    // now analyse the body

    do
    {
      QWKEKludge = false;

      if ((p = isQWKEKludge("To:", body, kline, QWKE_MAXFROMTO)))
      {
        body = p;
        lh->setTo(kline);
        QWKEKludge = true;

        if (isEmail)
        {
          na.set(kline);
          lh->setTo("INTERNET");
        }
      }

      if ((p = isQWKEKludge("From:", body, kline, QWKE_MAXFROMTO)))
      {
        body = p;
        lh->setFrom(kline);
        QWKEKludge = true;
      }

      if ((p = isQWKEKludge("Subject:", body, kline, QWKE_MAXSUBJECT)))
      {
        body = p;
        lh->setSubject(kline);
        QWKEKludge = true;
      }
    }
    while (QWKEKludge);
  }

  if (body != letterBody && *body == '\n') body++;

  return body;
}


inline void qwk::resetLetters ()
{
  NLID = 0;
}


void qwk::initMSF (int **msgstat)
{
  for (int a = (persArea ? 1 : 0); a < maxAreas; a++)
    for (int l = 0; l < areas[a].totmsgs; l++)
      msgstat[a][l] = 0;
}


int qwk::readMSF (int **msgstat)
{
  alldrv->XTIread(msgstat, (persArea ? 1 : 0));
  return (MSF_READ | MSF_REPLIED | MSF_PERSONAL | MSF_MARKED);
}


const char *qwk::saveMSF (int **msgstat)
{
  return alldrv->XTIsave(msgstat, (persArea ? 1 : 0), BBSid);
}


inline bool qwk::isLatin1 ()
{
  return false;
}


inline const char **qwk::getBulletins ()
{
  return bulletins;
}


bool qwk::readControl ()
{
  char *l, *p;

  if (!(ctrlFile = openFile("control.dat"))) return false;

  bm->resourceObject->set(BBSName, nextLine());     // 1: BBS name
  nextLine();                                       // 2: BBS city and state
  nextLine();                                       // 3: BBS phone number

  l = nextLine();                                   // 4: BBS Sysop name
  int slen = strlen(l);

  if (slen > 6)   // common format is: name, Sysop
  {
    p = l + slen - 5;

    if (strcasecmp(p, "Sysop") == 0)
    {
      if (*--p == ' ') p--;
      if (*p == ',') *p = '\0';
    }
  }

  bm->resourceObject->set(SysOpName, l);

  l = nextLine();                                   // 5: registration #,BBSid
  p = strtokr(l, ',');
  strncpy(BBSid, p, 8);
  BBSid[8] = '\0';

  nextLine();                                       // 6: packet creation time
  l = nextLine();                                   // 7: user name
  bm->resourceObject->set(LoginName, l);
  bm->resourceObject->set(AliasName, l);

  nextLine();                                       // 8: blank
  nextLine();                                       // 9: zero
  nextLine();                                       // 10: total # of messages
                                                    //     (not reliable)
  maxAreas = atoi(nextLine()) + (persArea ? 2 : 1); // 11: total # of confer-
                                                    //     ences minus 1
  areas = new QWK_AREA_INFO[maxAreas];

  for (int a = (persArea ? 1 : 0); a < maxAreas; a++)
  {
    areas[a].number = atoi(nextLine());             // 12: conference number
    sprintf(areas[a].areanum, "%d", areas[a].number);
    areas[a].title = strdupplus(nextLine());        // 13: conference name
    areas[a].flags = A_SUBSCRIBED | (qwke ? 0 : A_ALLOWPRIV);
  }

  if (persArea) areas[0].title = NULL;

  slen = sizeof(bulletinfiles[0]) - 1;

  for (int i = 0; i < 3; i++)
  {
    strncpy(bulletinfiles[i], nextLine(), slen);
    bulletinfiles[i][slen] = '\0';
  }

  bulletins = alldrv->bulletin_list(bulletinfiles, 3, BL_NEWFILES | BL_QWK);

  fclose(ctrlFile);
  return true;
}


FILE *qwk::openFile (const char *fname)
{
  FILE *file;

  if (!(file = bm->fileList->ftryopen(fname, "rb")))
    sprintf(error, "Could not open %s file.", fname);

  return file;
}


char *qwk::nextLine ()
{
  static char line[MYMAXLINE];

  if (!fgets(line, sizeof(line), ctrlFile)) *line = '\0';

  return mkstr(line);
}


bool qwk::readExtInformation ()
{
  char *line;
  int num;

  if (!(ctrlFile = openFile("toreader.ext"))) return false;

  while (!feof(ctrlFile))
  {
    line = nextLine();

    if (strncasecmp(line, "AREA ", 5) == 0)
    {
      if (sscanf(line + 5, "%d %s", &num, line) == 2)
      {
        if (strchr(line, 'F')) areas[getAreaIndex(num)].flags |= A_FORCED;
        if (strchr(line, 'N') ||
            strchr(line, 'I')) areas[getAreaIndex(num)].flags |= A_NETMAIL;
        if (strchr(line, 'I') ||
            strchr(line, 'U')) areas[getAreaIndex(num)].flags |= A_INTERNET;
        if (strchr(line, 'R')) areas[getAreaIndex(num)].flags |= A_READONLY;
        if (strchr(line, 'H')) areas[getAreaIndex(num)].flags |= A_ALIAS;
        if (strchr(line, 'P') ||
            strchr(line, 'X')) areas[getAreaIndex(num)].flags |= A_ALLOWPRIV;
      }
    }

    else if (strncasecmp(line, "ALIAS ", 6) == 0)
    {
      cropesp(line);
      bm->resourceObject->set(AliasName, line + 6);
    }
  }

  fclose(ctrlFile);
  return true;
}


bool qwk::readIndices ()
{
  int a, msgs;
  const char *fname;

  for (a = 0; a < maxAreas; a++)
  {
    body[a] = NULL;
    areas[a].totmsgs = 0;
  }

  bm->fileList->gotoFile(-1);

  if ((fname = bm->fileList->getNext(".ndx")))   // a first index file exists
  {
    char aname[9];
    FILE *ndxFile;
    struct
    {
      unsigned char recnum[4];
      unsigned char confnum;
    } ndx_rec;

    while (fname)
    {
      msgs = 0;
      int len = strlen(fname) - 4;

      strncpy(aname, fname, len);
      aname[len] = '\0';
      a = atoi(aname);

      if (a == 0)
      {
        if (strcasecmp(aname, "personal") == 0) a = -1;
        else if (strcmp(aname, "000") != 0) goto NEXTNDX;   // invalid name
      }

      a = getAreaIndex(a);

      if ((ndxFile = bm->fileList->ftryopen(fname, "rb")))
      {
        msgs = bm->fileList->getSize() / NDX_RECLEN;
        body[a] = new bodytype[msgs];

        for (int l = 0; l < msgs; l++)
        {
          if (fread(&ndx_rec, NDX_RECLEN, 1, ndxFile) != 1)
          {
            strcpy(error, "Could not read index.");
            fclose(ndxFile);
            return false;
          }

          body[a][l].pointer = getMKSlong(ndx_rec.recnum) << 7;
        }

        fclose(ndxFile);
        areas[a].totmsgs = msgs;
      }

NEXTNDX:
      fname = bm->fileList->getNext(".ndx");
    }
  }

  else   // build own index
  {
    MSG_NDX *anchor = new MSG_NDX, *ndx = anchor, *i, *next;
    qwk_header qh;

    const char *username = bm->resourceObject->get(LoginName);
    const char *aliasname = bm->resourceObject->get(AliasName);

    fseek(msgFile, MSGDAT_RECLEN, SEEK_SET);

    while (qh.read(msgFile))
    {
      if (!qh.skipit)
      {
        ndx->area = getAreaIndex(qh.confnum);
        ndx->pointer = ftell(msgFile);
        areas[ndx->area].totmsgs++;

        ndx->next = new MSG_NDX;
        ndx = ndx->next;

        if (persArea && (strcasecmp(qh.to, username) == 0 ||
                         strcasecmp(qh.to, aliasname) == 0))
        {
          ndx->area = 0;
          ndx->pointer = ftell(msgFile);
          areas[0].totmsgs++;

          ndx->next = new MSG_NDX;
          ndx = ndx->next;
        }

        fseek(msgFile, qh.msglen, SEEK_CUR);
      }
    }

    for (a = 0; a < maxAreas; a++)
    {
      body[a] = new bodytype[areas[a].totmsgs];
      areas[a].totmsgs = 0;
    }

    i = anchor;

    while (i != ndx)
    {
      next = i->next;

      a = i->area;
      body[a][areas[a].totmsgs++].pointer = i->pointer;

      delete i;
      i = next;
    }

    delete ndx;
  }

  return true;
}


int qwk::getAreaIndex (int area) const
{
  int a;

  // -1 is an internal pseudo-number for the personal area
  if (area == -1) return 0;

  for (a = (persArea ? 1 : 0); a < maxAreas; a++)
    if (areas[a].number == area) break;

  return (a == maxAreas ? -1 : a);
}


int qwk::getLetterIndex (int area, unsigned long pos) const
{
  int l;

  for (l = 0; l < areas[area].totmsgs; l++)
    if ((unsigned long) body[area][l].pointer == pos) break;

  return l;
}


inline const char *qwk::getBBSid () const
{
  return BBSid;
}


bool qwk::isQWKE () const
{
  return qwke;
}


char *qwk::isQWKEKludge (const char *kludge, char *source,
                         char *dest, int destlen) const
{
  int len = strlen(kludge), i;

  if (strncmp(source, kludge, len) == 0)
  {
    source += len;
    while (*source == ' ') source++;

    for (i = 0; *source && (*source != '\n') && (i < destlen); i++)
      dest[i] = *source++;
    dest[i] = '\0';

    while (*source && (*source != '\n')) source++;

    return (*source ? ++source : source);
  }
  else return NULL;
}


/*
   QWK / QWKE reply driver
*/

qwk_reply::qwk_reply (bmail *bm, main_driver *mainDriver)
{
  this->bm = bm;
  this->mainDriver = (qwk *) mainDriver;
  alldrv = new alldriver(bm, this);
  NLID = 1;
  noOfLetters = 0;
  replyBody = NULL;
  firstReply = currentReply = NULL;
}


qwk_reply::~qwk_reply ()
{
  REPLY *next, *r = firstReply;

  while (r)
  {
    next = r->next;
    if (r->filename) remove(r->filename);
    delete[] r->filename;
    delete[] r->address;
    delete r;
    r = next;
  }

  delete[] replyBody;
  delete alldrv;
}


const char *qwk_reply::init ()
{
  sprintf(replyPacketName, "%.8s.rep", this->mainDriver->getBBSid());

  if (bm->getServiceType() != ST_REPLY) replyInf();
  if (!getReplies(replyPacketName)) return error;

  return NULL;
}


inline int qwk_reply::getNoOfAreas ()
{
  return 1;
}


inline area_header *qwk_reply::getNextArea ()
{
  return getArea();
}


void qwk_reply::selectArea (int area)
{
  // REPLY_AREA has area number 0
  if (area == 0) resetLetters();
}


inline int qwk_reply::getNoOfLetters ()
{
  return noOfLetters;
}


letter_header *qwk_reply::getNextLetter ()
{
  int area = mainDriver->getAreaIndex(currentReply->qh.confnum);
  area += (area == -1 ? 1 : bm->driverList->getOffset(mainDriver));

  // this is for QWKE only
  net_address na;
  if (currentReply->address) na.set(currentReply->address);

  letter_header *newLetter = new letter_header(bm, NLID,
                                               currentReply->qh.from,
                                               currentReply->qh.to,
                                               currentReply->qh.subject,
                                               currentReply->qh.date,
                                               NLID,
                                               currentReply->qh.refnum,
                                               area,
                                               (currentReply->qh.privat ? PRIVATE : 0),
                                               isLatin1(),
                                               (currentReply->address ? &na : NULL),
                                               this);

  NLID++;
  currentReply = currentReply->next;

  return newLetter;
}


const char *qwk_reply::getBody (int area, int letter, letter_header *lh)
{
  FILE *replyFile;

  (void) area;
  delete[] replyBody;

  REPLY *r = firstReply;
  for (int l = 1; l < letter; l++) r = r->next;

  if ((replyFile = fopen(r->filename, "rt")))
  {
    long msglen = r->qh.msglen;
    replyBody = new char[msglen + 1];
    msglen = fread(replyBody, sizeof(char), msglen, replyFile);
    if (msglen && (replyBody[msglen - 1] == '\n')) msglen--;
    replyBody[msglen] = '\0';

    // will anyone need this, i.e. use a charset kludge in replies?
    bool latin1 = lh->isLatin1();
    alldrv->fsc0054(replyBody, '\1', &latin1);
    lh->setLatin1(latin1);

    fclose(replyFile);
  }
  else replyBody = strdupplus("");

  return replyBody;
}


void qwk_reply::resetLetters ()
{
  NLID = 1;
  currentReply = firstReply;
}


inline void qwk_reply::initMSF (int **msgstat)
{
  (void) msgstat;
}


inline int qwk_reply::readMSF (int **msgstat)
{
  (void) msgstat;
  return 0;
}


inline const char *qwk_reply::saveMSF (int **msgstat)
{
  (void) msgstat;
  return NULL;
}


inline bool qwk_reply::isLatin1 ()
{
  return false;
}


inline const char **qwk_reply::getBulletins ()
{
  return NULL;
}


void qwk_reply::enterLetter (letter_header *newLetter,
                             const char *newLetterFileName, long msglen)
{
  static char fmt[] = "%m-%d-%y %H:%M";     // to avoid warning for %y

  REPLY *newReply = new REPLY;
  memset(newReply, 0, sizeof(REPLY));

  strncpy(newReply->qh.from, newLetter->getFrom(), QWKE_MAXFROMTO);
  strncpy(newReply->qh.to, newLetter->getTo(), QWKE_MAXFROMTO);
  strncpy(newReply->qh.subject, newLetter->getSubject(), QWKE_MAXSUBJECT);

  time_t now = time(NULL);
  strftime(newReply->qh.date, sizeof(newReply->qh.date), fmt, localtime(&now));

  newReply->qh.refnum = newLetter->getReplyTo();
  newReply->qh.confnum = atoi(bm->areaList->getNumber());
  newReply->qh.msglen = msglen;
  newReply->qh.privat = newLetter->isPrivate();

  newReply->address =
    (bm->areaList->isNetmail() ? strdupplus(newLetter->getNetAddr().get(bm->areaList->isInternet()))
                               : NULL);

  newReply->filename = strdupplus(newLetterFileName);
  newReply->next = NULL;

  if (!firstReply) firstReply = newReply;

  if (noOfLetters)
  {
    REPLY *r = firstReply;
    // find last reply
    for (int l = 1; l < noOfLetters; l++) r = r->next;
    r->next = newReply;
  }

  noOfLetters++;
}


void qwk_reply::killLetter (int letter)
{
  REPLY *killReply, *r;

  if (noOfLetters == 0 || letter < 1 || letter > noOfLetters)
    fatalError("Could not kill letter.");

  if (letter == 1)
  {
    killReply = firstReply;
    firstReply = firstReply->next;
  }
  else
  {
    r = firstReply;
    for (int l = 1; l < letter - 1; l++) r = r->next;
    killReply = r->next;
    r->next = (letter == noOfLetters ? NULL : r->next->next);
  }

  remove(killReply->filename);
  delete[] killReply->filename;
  delete[] killReply->address;
  delete killReply;

  noOfLetters--;
  resetLetters();
}


const char *qwk_reply::newLetterBody (int letter, long size)
{
  REPLY *r = firstReply;
  for (int l = 1; l < letter; l++) r = r->next;

  if (size >= 0) r->qh.msglen = size;
  return r->filename;
}


area_header *qwk_reply::getArea ()
{
  return alldrv->getReplyArea((mainDriver->isQWKE() ? "QWKE Replies"
                                                    : "QWK Replies"),
                              noOfLetters);
}


void qwk_reply::replyInf ()
{
  char fname[MYMAXPATH];
  const char *dir = bm->resourceObject->get(InfWorkDir);
  const char *ctrl = "control.dat";
  const char *toreader = "toreader.ext";

  mkfname(fname, dir, ctrl);
  fcopy(ctrl, fname);

  if (mainDriver->isQWKE())
  {
    mkfname(fname, dir, toreader);
    fcopy(toreader, fname);
  }

  fcreate(dir, "messages.dat");

  bm->service->createSystemInfFile(fmtime(ctrl), replyPacketName);
}


bool qwk_reply::makeReply ()
{
  char rname[MYMAXPATH], sname[MYMAXPATH];
  bool success;

  mkfname(rname, bm->resourceObject->get(ReplyDir), replyPacketName);
  bool oldReplyPacket = (access(rname, R_OK | W_OK) == 0);

  if (noOfLetters == 0)
  {
    if (oldReplyPacket) return (remove(rname) == 0);
    else return true;
  }

  if (mychdir(bm->resourceObject->get(ReplyWorkDir)) != 0) return false;

  // rename old file, so don't trash old replies if packing fails
  strcpy(sname, rname);
  sname[strlen(sname) - 1] = '$';
  if (oldReplyPacket && (rename(rname, sname) == -1)) return false;

  if ((success = putReplies()))
  {
    // pack replies
    bm->service->pack(bm->resourceObject->get(ReplyDir), replyPacketName, "*.*");

    if ((success = (access(rname, R_OK | W_OK) == 0))) remove(sname);
  }

  if (!success) rename(sname, rname);

  clearDirectory(".");     // clean up
  return success;
}


bool qwk_reply::putReplies ()
{
  char msgFile[13];
  FILE *msg;
  char buf[MSGDAT_RECLEN];
  bool success = true;

  sprintf(msgFile, "%.8s.msg", mainDriver->getBBSid());

  if (!(msg = fopen(msgFile, "wb"))) return false;

  sprintf(buf, "%-128s", mainDriver->getBBSid());
  if (fwrite(buf, MSGDAT_RECLEN, 1, msg) != 1)
  {
    fclose(msg);
    return false;
  }

  REPLY *r = firstReply;

  for (int l = 0; l < noOfLetters; l++)
  {
    if (!put1Reply(msg, r))
    {
      success = false;
      break;
    }
    r = r->next;
  }

  fclose(msg);
  return success;
}


bool qwk_reply::put1Reply (FILE *msg, REPLY *r)
{
  bool isFido = false;
  net_address na;
  char *QWKEKludge = NULL;
  long extralen = 0, chunks;
  FILE *reply;
  char *replyText, *p, *oldsubj = NULL;
  bool success;

  if (mainDriver->isQWKE())
  {
    // netmail
    if (r->address)
    {
      na.set(r->address);
      isFido = (na.get() != NULL);

      if (!isFido)
      {
        strncpy(r->qh.to, na.get(true), QWKE_MAXFROMTO);
        r->qh.to[QWKE_MAXFROMTO] = '\0';
      }
    }

    long tolen = strlen(r->qh.to);
    bool tolong = ((tolen > 25) || (r->address && !isFido));

    long fromlen = strlen(r->qh.from);
    bool fromlong = (fromlen > 25);

    long subjlen = strlen(r->qh.subject);
    bool subjlong = ((subjlen > 25) || (r->address && isFido));

    if (tolong) extralen += 4 + tolen + 1;
    if (fromlong) extralen += 6 + fromlen + 1;
    if (subjlong) extralen += 9 + subjlen + 1;

    if (extralen > 0)
    {
      extralen++;
      QWKEKludge = new char[extralen + 1];
      *QWKEKludge = '\0';

      if (tolong)
      {
        strcat(QWKEKludge, "To: ");
        strcat(QWKEKludge, r->qh.to);
        strcat(QWKEKludge, "\n");
      }

      if (fromlong)
      {
        strcat(QWKEKludge, "From: ");
        strcat(QWKEKludge, r->qh.from);
        strcat(QWKEKludge, "\n");
      }

      if (subjlong)
      {
        strcat(QWKEKludge, "Subject: ");
        strcat(QWKEKludge, r->qh.subject);
        strcat(QWKEKludge, "\n");
      }

      strcat(QWKEKludge, "\n");
    }
  }

  if (r->address)
  {
    if (isFido)
    {
      oldsubj = strdupplus(r->qh.subject);
      sprintf(r->qh.subject, "@%s", na.get());
    }
    else strcpy(r->qh.to, "INTERNET");
  }

  r->qh.msglen += extralen;
  success = r->qh.write(msg);
  r->qh.msglen -= extralen;

  if (!success)
  {
    if (oldsubj) strcpy(r->qh.subject, oldsubj);
    delete[] oldsubj;
    return false;
  }

  if ((chunks = (r->qh.msglen + extralen + MSGDAT_RECLEN - 1) >> 7) == 0)
    return true;

  if (!(reply = fopen(r->filename, "rt"))) return false;

  replyText = new char[(chunks << 7) + 1];
  memset(&replyText[(chunks - 1) << 7], ' ', MSGDAT_RECLEN);

  if (QWKEKludge) strcpy(replyText, QWKEKludge);

  success = (fread(replyText + extralen, sizeof(char), r->qh.msglen, reply) ==
                                                       (size_t) r->qh.msglen);
  fclose(reply);

  if (success)
  {
    replyText[r->qh.msglen + extralen] = '\0';

    for (p = replyText; *p; p++)
      if (*p == '\n') *p = (char) 227;

    p--;
    replyText[r->qh.msglen + extralen] = (*p == (char) 227 ? ' ' : (char) 227);

    success = (fwrite(replyText, MSGDAT_RECLEN, chunks, msg) ==
                                                             (size_t) chunks);
  }

  delete[] oldsubj;
  delete[] QWKEKludge;
  delete[] replyText;
  return success;
}


bool qwk_reply::getReplies (const char *replyPacketName)
{
  file_list *replies;
  const char *msgFile = NULL;
  FILE *msg = NULL;
  bool success = true;

  if (!(replies = alldrv->unpackReplies(replyPacketName))) return true;

  if (!(msgFile = replies->exists(".msg")) ||
      !(msg = replies->ftryopen(msgFile, "rb")))
  {
    strcpy(error, "Could not open reply file.");
    success = false;
  }
  else
  {
    off_t filesize = fsize(msgFile);
    fseek(msg, MSGDAT_RECLEN, SEEK_SET);     // skip header

    REPLY seed, *r = &seed;
    seed.next = NULL;

    while (ftell(msg) < filesize)
    {
      r->next = new REPLY;
      r = r->next;
      r->filename = NULL;
      r->address = NULL;
      r->next = NULL;

      if (!get1Reply(msg, r))
      {
        strcpy(error, "Could not get reply.");
        success = false;
        break;
      }

      noOfLetters++;
    }

    firstReply = seed.next;
  }

  if (msg) fclose(msg);
  if (msgFile) remove(msgFile);
  delete replies;

  return success;
}


bool qwk_reply::get1Reply (FILE *msg, REPLY *r)
{
  FILE *reply;
  char filename[13], filepath[MYMAXPATH];
  static long fnr = 1;
  char *replyText, *p;
  bool success;

  if (!r->qh.read(msg)) return false;

  sprintf(filename, "%08ld", fnr++);
  mkfname(filepath, bm->resourceObject->get(WorkDir), filename);

  if (!(reply = fopen(filepath, "wt"))) return false;

  replyText = new char[r->qh.msglen + 1];
  success = (fread(replyText, sizeof(char), r->qh.msglen, msg) ==
                                                       (size_t) r->qh.msglen);

  if (success)
  {
    p = &replyText[r->qh.msglen - 1];
    while (p > replyText && (*p == ' ' || *p == (char) 227)) p--;
    *++p = '\0';

    for (p = replyText; *p; p++)
      if (*p == (char) 227) *p = '\n';

    *p = '\n';
    r->qh.msglen = p - replyText + 1;

    char *text = replyText;

    if (mainDriver->isQWKE())
    {
      char hfield[QWKE_MAXSUBJECT + 1];
      bool QWKEKludge;
      char *p;

      // try to figure out whether it may be netmail

      strcpy(hfield, r->qh.to);
      cropesp(hfield);
      bool isEmail = (strcasecmp(hfield, "INTERNET") == 0);

      strcpy(hfield, r->qh.subject);
      cropesp(hfield);
      if (*hfield == '@') r->address = strdupplus(hfield + 1);

      // now analyse the body

      do
      {
        QWKEKludge = false;

        if ((p = mainDriver->isQWKEKludge("To:", text, r->qh.to,
                                                       QWKE_MAXFROMTO)))
        {
          r->qh.msglen -= p - text;
          text = p;
          QWKEKludge = true;

          if (isEmail)
          {
            r->address = strdupplus(r->qh.to);
            strcpy(r->qh.to, "INTERNET");
          }
        }

        if ((p = mainDriver->isQWKEKludge("From:", text, r->qh.from,
                                                         QWKE_MAXFROMTO)))
        {
          r->qh.msglen -= p - text;
          text = p;
          QWKEKludge = true;
        }

        if ((p = mainDriver->isQWKEKludge("Subject:", text, r->qh.subject,
                                                            QWKE_MAXSUBJECT)))
        {
          r->qh.msglen -= p - text;
          text = p;
          QWKEKludge = true;
        }
      }
      while (QWKEKludge);
    }

    if (text != replyText && *text == '\n')
    {
      text++;
      r->qh.msglen--;
    }

    success = (fwrite(text, sizeof(char), r->qh.msglen, reply) ==
                                                       (size_t) r->qh.msglen);
    r->filename = strdupplus(filepath);
  }

  fclose(reply);
  delete[] replyText;
  return success;
}
