/*
 * blueMail offline mail reader
 * SOUP driver

 Copyright (c) 2002 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */


#include <ctype.h>
#include <string.h>
#include "soup.h"
#include "../../common/auxil.h"
#include "../../common/error.h"
#include "../../common/mysystem.h"


/*
   SOUP main driver
*/

soup::soup (bmail *bm)
{
  this->bm = bm;
  alldrv = new alldriver(bm, this);

  NAID = 0;   // next area ID
              // counts the areas already returned by getNextArea()
  NLID = 0;   // next letter ID
              // counts the letters already returned by getNextLetter()
  currentArea = 0;
  maxAreas = 0;

  areas = NULL;
  headerlines = NULL;

  body = NULL;
  letterBody = NULL;
  bulletins = NULL;
  areasFile = msgFile = NULL;
}


soup::~soup ()
{
  while (maxAreas--)
  {
    delete[] body[maxAreas];
    delete[] headerlines[maxAreas];
    delete[] areas[maxAreas]->prefix;
    delete areas[maxAreas];
  }

  delete[] bulletins;
  delete[] letterBody;
  delete[] body;
  delete[] headerlines;
  delete[] areas;
  delete alldrv;

  if (msgFile) fclose(msgFile);
  if (areasFile) fclose(areasFile);
}


const char *soup::init ()
{
  const char *username = bm->resourceObject->get(UserName);
  bm->resourceObject->set(LoginName, username);
  bm->resourceObject->set(AliasName, username);
  bm->resourceObject->set(SysOpName, username);
  bm->resourceObject->set(BBSName, "Usenet offline");

  bool areas_ok = getAreas();

  body = new bodytype *[maxAreas];
  headerlines = new long *[maxAreas];
  for (int i = 0; i < maxAreas; i++)
  {
    body[i] = NULL;
    headerlines[i] = NULL;
  }

  if (!areas_ok || !buildIndices()) return error;

  const char bulletinfile[2][13] = {"info", "list"};
  bulletins = alldrv->bulletin_list(bulletinfile, 2, BL_NOADD);

  return NULL;
}


inline int soup::getNoOfAreas ()
{
  return maxAreas;
}


area_header *soup::getNextArea ()
{
  area_header *ah;

  int totmsgs = areas[NAID]->totmsgs;
  int numpers = (areas[NAID]->isEmail ? totmsgs : 0);

  int flags = A_SUBSCRIBED | A_INTERNET |
              (areas[NAID]->isEmail ? A_NETMAIL : 0);

  fseek(areasFile, areas[NAID]->offset, SEEK_SET);
  fgetsnl(line, MAXLINELEN, areasFile);
  strtok(line, "\t");
  char *areaname = strtok(NULL, "\t");

  ah = new area_header(bm, NAID + bm->driverList->getOffset(this),
                       areas[NAID]->number, areaname, areaname, "SOUP",
                       flags, totmsgs, numpers, 80, OTHERADDRLEN);
  NAID++;

  return ah;
}


void soup::selectArea (int area)
{
  currentArea = area;
  resetLetters();
}


inline int soup::getNoOfLetters ()
{
  return areas[currentArea]->totmsgs;
}


letter_header *soup::getNextLetter ()
{
  net_address na;
  bool latin1 = isLatin1();
  char *from = NULL, *faddr = NULL, *to = NULL, *subj = NULL, *date = NULL;
  int name, n_len, addr, a_len;

  bool c_news = (areas[currentArea]->format == 'i');

  if (!openFile(areas[currentArea])) fatalError(error);
  fseek(msgFile, body[currentArea][NLID].pointer, SEEK_SET);

  if (c_news)
  {
    fgetsnl(line, MAXLINELEN, msgFile);
    subj = strdupplus(strtokn(line, '\t'));
    char *froml = strtokn(NULL, '\t');
    date = strdupplus(strtokn(NULL, '\t'));

    alldrv->splitAddress(froml, name, n_len, addr, a_len);

    if (n_len)
    {
      from = new char[n_len + 1];
      strncpy(from, froml + name, n_len);
      from[n_len] = '\0';
    }

    if (a_len)
    {
      faddr = new char[a_len + 1];
      strncpy(faddr, froml + addr, a_len);
      faddr[a_len] = '\0';
    }
  }
  else
  {
    headerlines[currentArea][NLID] =
      alldrv->parseHeader(msgFile, body[currentArea][NLID].pointer +
                                   body[currentArea][NLID].msglen, &latin1);

    from = alldrv->getFrom();
    faddr = alldrv->getFaddr();
    to = alldrv->getTo();
    subj = alldrv->getSubject();
    date = alldrv->getDate();
  }

  if (faddr) na.set(faddr);

  letter_header *lh = new letter_header(bm, NLID,
                                        (from ? from : ""),
                                        (to ? to : ""),
                                        (subj ? subj : ""),
                                        (date ? date : ""),
                                        NLID + 1, 0, currentArea,
                                        (areas[currentArea]->isEmail ? PRIVATE
                                                                     : 0),
                                        latin1, &na, this);

  if (c_news)
  {
    delete[] from;
    delete[] faddr;
    delete[] subj;
    delete[] date;
  }

  NLID++;
  return lh;
}


const char *soup::getBody (int area, int letter, letter_header *lh)
{
  (void) lh;
  int kar;

  // get number of headerlines plus separation line
  long hlines = headerlines[area][letter] + 1;

  if (areas[area]->format == 'i') return "(no body available)";

  delete[] letterBody;
  letterBody = new char[body[area][letter].msglen + hlines + 1];
  unsigned char *p = (unsigned char *) letterBody;

  if (!openFile(areas[area])) fatalError(error);
  fseek(msgFile, body[area][letter].pointer, SEEK_SET);

  bool startOfLine = true;

  for (long c = 0; c < body[area][letter].msglen; c++)
  {
    kar = fgetc(msgFile);

    // a message shouldn't (must not) contain a 0x00 character
    if (kar == 0) kar = ' ';

    // convert header lines into kludge lines
    if (startOfLine && hlines)     // to prevent letterBody from overflow
    {
      *p++ = '\1';
      hlines--;
    }

    *p++ = kar;

    startOfLine = (kar == '\n');
  }

  do p--;
  // strip blank lines
  while (p >= (unsigned char *) letterBody && (*p == ' ' || *p == '\n'));

  *++p = '\0';

  // decode quoted-printable
  if (bm->driverList->allowQuotedPrintable() && alldrv->isQP(letterBody))
    mimeDecodeBody(letterBody);

  return letterBody;
}


inline void soup::resetLetters ()
{
  NLID = 0;
}


void soup::initMSF (int **msgstat)
{
  for (int a = 0; a < maxAreas; a++)
    for (int l = 0; l < areas[a]->totmsgs; l++)
      msgstat[a][l] = 0;
}


int soup::readMSF (int **msgstat)
{
  alldrv->XTIread(msgstat, 0, "XTI");
  return (MSF_READ | MSF_REPLIED | MSF_MARKED);
}


const char *soup::saveMSF (int **msgstat)
{
  return alldrv->XTIsave(msgstat, 0, "XTI", false);
}


inline bool soup::isLatin1 ()
{
  return true;
}


inline const char **soup::getBulletins ()
{
  return bulletins;
}


bool soup::getAreas ()
{
  char *prefix, *name, *encoding;
  SOUP_AREA_INFO anchor, *m = &anchor;
  bool success = true;

  if ((areasFile = bm->fileList->ftryopen("areas", "rt")))
  {
    int totareas = 0;
    long lastpos = 0;

    while (fgetsnl(line, MAXLINELEN, areasFile))
    {
      mkstr(line);
      prefix = strtok(line, "\t");
      name = strtok(NULL, "\t");
      encoding = strtok(NULL, "\t");

      if (prefix && name && encoding && (strlen(encoding) >= 2))
      {
        totareas++;
        m->next = new SOUP_AREA_INFO;
        m = m->next;
        sprintf(m->number, "%5d", totareas);
        m->prefix = strdupplus(prefix);
        m->format = encoding[0];
        m->index = encoding[1];
        m->isEmail = ((encoding[2] == 'm') ||
                      ((encoding[2] != 'n') && ((*encoding == 'm') ||
                                                (*encoding == 'M') ||
                                                (*encoding == 'b'))));
        m->totmsgs = 0;
        m->offset = lastpos;
        m->next = NULL;

        if (!openFile(m))
        {
          success = false;
          break;
        }
      }

      lastpos = ftell(areasFile);
    }

    if (totareas)
    {
      areas = new SOUP_AREA_INFO *[totareas];
      m = anchor.next;
      int i = 0;

      while (m)
      {
        areas[i++] = m;
        m = m->next;
      }

      maxAreas = totareas;
    }
  }
  else
  {
    strcpy(error, "Could not open areas file.");
    success = false;
  }

  return success;
}


// if there is no index file, we have to build the indices anyway, so it's
// (probably?) not worth using existing 'c', 'C' or 'i' format index files
bool soup::buildIndices ()
{
  int totmsgs;
  bool binary, c_news;
  size_t readlen = 0;
  long msglen, pos, lpos;
  unsigned char bdata[4];
  bool success = true;

  struct MSG_IDX
  {
    long offset;
    long msglen;
    MSG_IDX *next;
  } anchor, *ndx, *d;

  for (int a = 0; a < maxAreas; a++)
  {
    totmsgs = 0;
    ndx = &anchor;
    binary = (toupper(areas[a]->format) == 'B');
    c_news = (toupper(areas[a]->index) == 'C');

    if (openFile(areas[a]))
    {
      while (binary ? (fread(bdata, sizeof(char), 4, msgFile) == 4)
                    : (readlen = fgetsnl(line, MAXLINELEN, msgFile)))
      {
        msglen = -1;
        pos = ftell(msgFile);

        switch (areas[a]->format)
        {
          case 'u':
            sscanf(line, "#! rnews %ld", &msglen);
            if (msglen == 0) msglen = -1;            // ignore it
            break;

          case 'm':
            if (strncmp(line, "From ", 5) == 0)
            {
              do lpos = ftell(msgFile);
              while (fgetsnl(line, MAXLINELEN, msgFile) &&
                     strncmp(line, "From ", 5) != 0);

              msglen = lpos - pos;
              fseek(msgFile, pos, SEEK_SET);     // back to behind "From "
            }
            break;

          case 'M':
            // we won't handle that format
            break;

          case 'b':
          case 'B':
            msglen = getLSBlong(bdata);
            break;

          case 'i':
            if (c_news)
            {
              pos -= readlen - strlen(strtokn(line, '\t')) - 1;
              msglen = 0;
            }
            break;
        }

        if (msglen != -1)
        {
          totmsgs++;
          ndx->next = new MSG_IDX;
          ndx = ndx->next;
          ndx->offset = pos;
          ndx->msglen = msglen;
          ndx->next = NULL;

          fseek(msgFile, msglen, SEEK_CUR);
        }
      }
    }
    else success = false;

    if (totmsgs)
    {
      body[a] = new bodytype[totmsgs];
      headerlines[a] = new long[totmsgs];
      ndx = anchor.next;
      int i = 0;

      while (ndx)
      {
        body[a][i].pointer = ndx->offset;
        body[a][i].msglen = ndx->msglen;
        headerlines[a][i] = 0;               // will be set in getNextLetter()
        i++;
        d = ndx;
        ndx = ndx->next;
        delete d;
      }

      areas[a]->totmsgs = totmsgs;
    }

    if (!success) break;
  }

  return success;
}


bool soup::openFile (SOUP_AREA_INFO *area)
{
  static long lastOffset;

  char fname[MYMAXPATH];
  const char *ext = (area->format == 'i' ? "IDX" : "MSG");
  const char *access;
  bool success = true;

  if (!msgFile) lastOffset = -1;     // initialize static variable

  if (area->offset != lastOffset)
  {
    lastOffset = area->offset;

    if (msgFile) fclose(msgFile);
    msgFile = NULL;

    mkfnamext(fname, area->prefix, ext);

    if (bm->fileList->exists(fname))
    {
      access = "open";
      msgFile = bm->fileList->ftryopen(fname, "rb");
    }
    else access = "find";

    if (!msgFile)
    {
      sprintf(error, "Could not %s %.20s.%s file.", access, area->prefix, ext);
      success = false;
    }
  }

  return success;
}


/*
   SOUP reply driver
*/

//void soup_reply::replyInf ()
//{
//  char fname[MYMAXPATH], *ext;
//  const char *areas = bm->fileList->exists("areas");
//  const char *dir = bm->resourceObject->get(InfWorkDir);
//
//  mkfname(fname, dir, "AREAS");
//  fcopy(areas, fname);
//
//  // to be independent from mainDriver,
//  // we create the area files by using fileList
//  for (int i = 0; i < bm->fileList->getNoOfFiles(); i++)
//  {
//    bm->fileList->gotoFile(i);
//    ext = strrchr(bm->fileList->getName(), '.');
//
//    if (ext && (strcasecmp(ext, ".MSG") == 0 ||
//                strcasecmp(ext, ".IDX") == 0))
//      fcreate(dir, bm->fileList->getName());
//  }
//
//  bm->service->createSystemInfFile(fmtime(areas), replyPacketName);
//}
