/*
 * blueMail offline mail reader
 * Unix mail file (Berkeley format - mbox) driver

 Copyright (c) 2002 Ingo Brueckl <ib@wupperonline.de>

 Distributed under the GNU General Public License.
 For details, see the file COPYING in the parent directory. */


#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include "unixmail.h"
#include "../../common/auxil.h"
#include "../../common/error.h"


/*
   Unix mail file (Berkeley format - mbox) main driver
*/

unixmail::unixmail (bmail *bm)
{
  this->bm = bm;
  alldrv = new alldriver(bm, this);

  NLID = 0;   // next letter ID
              // counts the letters already returned by getNextLetter()
  totmsgs = 0;

  anchor = NULL;
  msg = NULL;

  letterBody = NULL;
  mbox = NULL;
}


unixmail::~unixmail ()
{
  while (totmsgs) delete msg[--totmsgs];

  delete[] letterBody;
  delete[] msg;
  delete alldrv;

  if (mbox) fclose(mbox);
}


const char *unixmail::init ()
{
  const char *username = bm->resourceObject->get(UserName);
  bm->resourceObject->set(LoginName, username);
  bm->resourceObject->set(AliasName, username);
  bm->resourceObject->set(SysOpName, "You");
  bm->resourceObject->set(BBSName, "Unix Mail File");

  mfile = bm->resourceObject->get(UnixMail);

  strncpy(lockfile, mfile, MYMAXPATH);
  lockfile[MYMAXPATH] = '\0';
  strcat(lockfile, ".lock");

  if (!(mbox = bm->fileList->ftryopen(fname(mfile), "xrt")) ||
      (access(lockfile, F_OK) == 0)) return "Could not open mail file.";

  buildIndices();

  return NULL;
}


inline int unixmail::getNoOfAreas ()
{
  return 1;
}


area_header *unixmail::getNextArea ()
{
  return new area_header(bm, bm->driverList->getOffset(this),
                         "PERS", "PERSONAL", "Letters addressed to you",
                         "Unix",
                         A_SUBSCRIBED | A_FORCED | A_NETMAIL | A_INTERNET,
                         totmsgs, totmsgs, 80, OTHERADDRLEN);
}


inline void unixmail::selectArea (int area)
{
  (void) area;
  resetLetters();
}


inline int unixmail::getNoOfLetters ()
{
  return totmsgs;
}


letter_header *unixmail::getNextLetter ()
{
  net_address na;
  bool latin1 = isLatin1();

  fseek(mbox, msg[NLID]->header, SEEK_SET);
  msg[NLID]->headerlines = alldrv->parseHeader(mbox, msg[NLID]->end, &latin1);

  if (alldrv->getFaddr()) na.set(alldrv->getFaddr());

  int letter = NLID++;
  return new letter_header(bm, letter,
                           alldrv->getFrom(),
                           alldrv->getTo(),
                           alldrv->getSubject(),
                           alldrv->getDate(),
                           NLID, 0, 0, PRIVATE, latin1, &na, this);
}


const char *unixmail::getBody (int area, int letter, letter_header *lh)
{
  (void) area;
  (void) lh;
  int kar;

  // get number of headerlines plus separation line
  long hlines = msg[letter]->headerlines + 1;

  delete[] letterBody;
  letterBody = new char[msg[letter]->end - msg[letter]->header + hlines + 1];
  unsigned char *p = (unsigned char *) letterBody;

  fseek(mbox, msg[letter]->header, SEEK_SET);

  bool startOfLine = true;

  while (ftell(mbox) < msg[letter]->end)
  {
    kar = fgetc(mbox);

    // a message shouldn't (must not) contain a 0x00 character
    if (kar == 0) kar = ' ';

    // convert header lines into kludge lines
    if (startOfLine && hlines)     // to prevent letterBody from overflow
    {
      *p++ = '\1';
      hlines--;
    }

    *p++ = kar;

    startOfLine = (kar == '\n');
  }

  do p--;
  // strip blank lines
  while (p >= (unsigned char *) letterBody && (*p == ' ' || *p == '\n'));

  *++p = '\0';

  // decode quoted-printable
  if (bm->driverList->allowQuotedPrintable() && alldrv->isQP(letterBody))
    mimeDecodeBody(letterBody);

  return letterBody;
}


inline void unixmail::resetLetters ()
{
  NLID = 0;
}


inline void unixmail::initMSF (int **msgstat)
{
  (void) msgstat;
}


int unixmail::readMSF (int **msgstat)
{
  for (int l = 0; l < totmsgs; l++) msgstat[0][l] = msg[l]->stat;

  return (MSF_READ | MSF_REPLIED | MSF_MARKED);
}


const char *unixmail::saveMSF (int **msgstat)
{
  // Is it possible to avoid copying the whole mail file twice without
  // extensive use of memory? After inserting the (X-)Status:-line(s),
  // a message can shrink or grow, so if we'd intend to do all the
  // writing only in the mail file, where should we start - at the top or
  // at the end? Another problem: OSes using CRLF, where the length of a
  // string read or written differs from its size in the file!
  // If anyone knows a better (i.e. quicker) solution, please let me know.

  int fd;
  FILE *tfile;
  size_t len;

  // create old style lockfile
  if ((fd = open(lockfile, O_WRONLY | O_CREAT | O_EXCL, 0666)) == -1)
  {
    if (errno == EEXIST) return NULL;
  }
  else close(fd);

  // take the mail file's fingerprint
  off_t msize = bm->fileList->getSize();
  time_t mtime = bm->fileList->getDate();
  fclose(mbox);

  // reopen it
  if (!(mbox = bm->fileList->ftryopen(fname(mfile), "Xr+t")) ||
      fsize(mfile) != msize || fmtime(mfile) != mtime)
  {
    remove(lockfile);
    return "!";           // no chance to handle this reopen error
  }

  // we now assume that the mail file hasn't been altered in the meantime

  const char *tmpfile = mymktmpfile(NULL);

  // recreate it temporarily
  if ((tfile = fopen(tmpfile, "w+t")))
  {
    int i = 0, n = 1;
    long pos, lastpos = -1;
    bool newline = true;

    while (fgets(line, MAXLINELEN, mbox))
    {
      pos = ftell(mbox);

      if (newline && (pos == msg[i]->body && msg[i]->body > msg[i]->header))
        fprintf(tfile, "%s", Status(msgstat[0][i]));

      if (lastpos != msg[i]->status && lastpos != msg[i]->x_status)
        fprintf(tfile, "%s", line);

      if (n < totmsgs && pos == msg[n]->top) i = n++;

      newline = (strchr(line, '\n') != NULL);
      if (newline) lastpos = pos;
    }

    if (fseek(mbox, 0, SEEK_SET) != 0 || fseek(tfile, 0, SEEK_SET) != 0)
    {
      fclose(tfile);
      remove(tmpfile);
      remove(lockfile);
      return NULL;
    }

    while ((len = fread(line, sizeof(char), MAXLINELEN, tfile)))
      if (fwrite(line, sizeof(char), len, mbox) != len)
      {
        fclose(tfile);
        catastrophe(true, tmpfile);
      }

    fclose(tfile);
    off_t tsize = fsize(tmpfile);

    if (tsize < msize)
      if (fseek(mbox, 0, SEEK_SET) != 0 ||
          ftruncate(fileno(mbox), tsize) != 0) catastrophe(false, tmpfile);

    remove(tmpfile);
    remove(lockfile);
    return "";
  }

  remove(tmpfile);
  remove(lockfile);
  return NULL;
}


inline bool unixmail::isLatin1 ()
{
  return true;
}


inline const char **unixmail::getBulletins ()
{
  return NULL;
}


void unixmail::buildIndices ()
{
  long lastpos = 0;
  bool isHeader = false;
  MSG_INFO *m = NULL, *last = NULL;

  while (fgetsnl(line, MAXLINELEN, mbox))
  {
    if (strncmp(line, "From ", 5) == 0)
    {
      totmsgs++;
      m = new MSG_INFO;
      if (!anchor) anchor = m;
      if (last) last->next = m;
      m->top = lastpos;
      m->header = ftell(mbox);
      m->status = 0;
      m->x_status = 0;
      m->body = m->header;     // just in case we won't find a body later
      m->stat = 0;
      m->headerlines = 0;      // will be set in getNextLetter()
      m->next = NULL;
      last = m;
      isHeader = true;
    }

    else if (!m) continue;     // junk before first "From "

    else if (isHeader && (strncasecmp(line, "Status:", 7) == 0 ||
                          strncasecmp(line, "X-Status:", 9) == 0))
    {
      if (line[1] == '-') m->x_status = lastpos;
      else m->status = lastpos;

      m->stat |= (strchr(line, 'R') ? MSF_READ : 0) |
                 (strchr(line, 'A') ? MSF_REPLIED : 0) |
                 (strchr(line, 'F') ? MSF_MARKED : 0);
    }

    else if (*line == '\n' && m->body == m->header)
    {
      m->body = ftell(mbox);
      isHeader = false;
    }

    lastpos = ftell(mbox);
  }

  if (totmsgs)
  {
    msg = new MSG_INFO *[totmsgs];
    m = anchor;
    int i = 0;
    while (m)
    {
      m->end = (m->next ? m->next->top : ftell(mbox));
      if (m->body == m->header) m->body = m->end;          // no body found
      msg[i++] = m;
      m = m->next;
    }
  }
}


const char *unixmail::Status (int stat) const
{
  static char statlines[] = "Status: _O\nX-Status: __\n";

  statlines[8] = (stat & MSF_READ ? 'R' : ' ');
  statlines[21] = (stat & MSF_REPLIED ? 'A' : ' ');
  statlines[22] = (stat & MSF_MARKED ? 'F' : ' ');

  return statlines;
}


void unixmail::catastrophe (bool mbox_destroyed, const char *tmpfile) const
{
  char message[131 + MYMAXPATH];

  sprintf(message, "%s%s%s%s",
                   "Could not rewrite mail file. ",
                   (mbox_destroyed ? "It certainly is destroyed now!" :
                                     "It contains junk at its end now!"),
                   "\n       A copy of the original mail file has been "
                                                   "stored as:\n\n       ",
                   canonize(tmpfile));
  fatalError(message);
}


/*
   Unix mail file (Berkeley format - mbox) reply driver
*/

//void unixmail_reply::replyInf ()
//{
//  fcreate(bm->resourceObject->get(InfWorkDir), fname(mfile));
//
//  // fmtime(mfile) doesn't seem to make sense as its fmtime constantly
//  // changes and we're creating the same system inf file over and over
//  // again, but we must keep in mind that its name may change, too
//  bm->service->createSystemInfFile(fmtime(mfile), replyPacketName);
//}
